import logging

import pygame

from trosnoth.const import MAP_TO_SCREEN_SCALE
from trosnoth.gui.framework import framework
from trosnoth.gui.framework.elements import TextElement, PictureElement
from trosnoth.gui.common import (Location, FullScreenAttachedPoint,
        ScaledSize, Area, CanvasX, Screen, RelativePoint)
from trosnoth.gui.framework.unobtrusiveValueGetter import YesNoGetter

from trosnoth.trosnothgui.ingame.messagebank import MessageBank
from trosnoth.trosnothgui.ingame.stars import StarGroup
from trosnoth.trosnothgui.ingame import mainMenu
from trosnoth.trosnothgui.ingame.gamevote import GameVoteMenu, NicknameBox
from trosnoth.trosnothgui.ingame.gauges import (TurretGauge, RespawnGauge,
        UpgradeGauge, GunGauge, StarGauge)
from trosnoth.trosnothgui.ingame.achievementBox import AchievementBox
from trosnoth.trosnothgui.ingame.chatBox import ChatBox
from trosnoth.trosnothgui.ingame.utils import mapPosToScreen
from trosnoth.gui.framework.dialogbox import DialogResult
from trosnoth.trosnothgui.settings.settings import SettingsMenu

from trosnoth.model.upgrades import (allUpgrades, Grenade, Ninja, Turret,
        Directatorship, Ricochet, Shoxwave)

from trosnoth.messages import (BuyUpgradeMsg, AbandonUpgradeMsg,
        RespawnRequestMsg, ChangeNicknameMsg, PlayerIsReadyMsg,
        ThrowTrosballMsg)

log = logging.getLogger('detailsInterface')


class DetailsInterface(framework.CompoundElement):
    '''Interface containing all the overlays onto the screen:
    chat messages, player lists, gauges, stars, etc.'''
    def __init__(self, app, gameInterface):
        super(DetailsInterface, self).__init__(app)
        self.gameInterface = gameInterface

        # Maximum number of messages viewable at any one time
        maxView = 8

        self.world = gameInterface.world
        self.player = None
        font = app.screenManager.fonts.messageFont
        self.currentMessages = MessageBank(self.app, maxView, 50,
                Location(FullScreenAttachedPoint(ScaledSize(-40,-40),
                'bottomright'), 'bottomright'), 'right', 'bottom', font)

        # If we want to keep a record of all messages and their senders
        self.input = None
        self.inputText = None
        self.unobtrusiveGetter = None
        self.turretGauge = None
        self.reloadGauge = GunGauge(self.app, Area(
                FullScreenAttachedPoint(ScaledSize(0,-60), 'midbottom'),
                ScaledSize(100,30), 'midbottom'))
        self.respawnGauge = None
        self.upgradeGauge = None
        self.achievementBox = None
        self.currentUpgrade = None
        self.starGroup = StarGroup(self.app)
        self.settingsMenu = SettingsMenu(app, onClose=self.hideSettings,
                showThemes=False)

        self.chatBox = ChatBox(app, self.world, self.gameInterface)

        menuloc = Location(FullScreenAttachedPoint((0,0), 'bottomleft'),
                'bottomleft')
        self.menuManager = mainMenu.MainMenu(self.app, menuloc, self,
                self.gameInterface.keyMapping)
        self.upgradeDisplay = UpgradeDisplay(app)
        self.trajectoryOverlay = TrajectoryOverlay(app, gameInterface.gameViewer.viewManager, self.upgradeDisplay)
        self.gameVoteMenu = GameVoteMenu(
            app, self.world, onChange=self._castGameVote)
        self._gameVoteUpdateCounter = 0
        self.elements = [
            self.currentMessages, self.upgradeDisplay, self.reloadGauge,
            self.starGroup, self.gameVoteMenu, self.chatBox,
            self.trajectoryOverlay, self.menuManager,
        ]

        self.upgradeMap = dict((upgradeClass.action, upgradeClass) for
                upgradeClass in allUpgrades)

    def sendRequest(self, msg):
        return self.gameInterface.sendRequest(msg)

    def tick(self, deltaT):
        super(DetailsInterface, self).tick(deltaT)
        if self.player is None:
            self.gameVoteMenu.active = False
            return
        else:
            if self._gameVoteUpdateCounter <= 0:
                self.gameVoteMenu.update(self.player)
                self._gameVoteUpdateCounter = 25
            else:
                self._gameVoteUpdateCounter -= 1

        self._updateTurretGauge()
        self._updateRespawnGauge()
        self._updateUpgradeGauge()
        self._updateAchievementBox()
        self._updateGameVoteMenu()

    def _castGameVote(self, msg):
        self.sendRequest(msg)

    def _updateGameVoteMenu(self):
        if self.world.canVote():
            self.gameVoteMenu.active = True
        else:
            self.gameVoteMenu.active = False

    def localAchievement(self, achievementId):
        if self.achievementBox is None:
            self.achievementBox = AchievementBox(self.app, self.player,
                                                 achievementId)
            self.elements.append(self.achievementBox)
        else:
            self.achievementBox.addAchievement(achievementId)

    def _updateAchievementBox(self):
        if (self.achievementBox is not None and
                len(self.achievementBox.achievements) == 0):
            self.elements.remove(self.achievementBox)
            self.achievementBox = None

    def _updateTurretGauge(self):
        player = self.player
        if self.turretGauge is None:
            if player.turret:
                self.turretGauge = TurretGauge(self.app, Area(
                        FullScreenAttachedPoint(ScaledSize(0,-100),
                        'midbottom'), ScaledSize(100,30), 'midbottom'),
                        player)
                self.elements.append(self.turretGauge)
        elif not player.turret:
            self.elements.remove(self.turretGauge)
            self.turretGauge = None

    def _updateRespawnGauge(self):
        player = self.player
        if self.respawnGauge is None:
            if player.dead:
                self.respawnGauge = RespawnGauge(self.app, Area(
                        FullScreenAttachedPoint(ScaledSize(0,-20),
                        'midbottom'), ScaledSize(100,30), 'midbottom'),
                        player, self.world)
                self.elements.append(self.respawnGauge)
        elif not player.dead:
            self.elements.remove(self.respawnGauge)
            self.respawnGauge = None

    def _updateUpgradeGauge(self):
        player = self.player
        if self.upgradeGauge is None:
            if player.upgrade is not None:
                self.upgradeGauge = UpgradeGauge(self.app, Area(
                        FullScreenAttachedPoint(ScaledSize(0,-20),
                        'midbottom'), ScaledSize(100,30), 'midbottom'),
                        player)
                self.elements.append(self.upgradeGauge)
        elif player.upgrade is None:
            self.elements.remove(self.upgradeGauge)
            self.upgradeGauge = None

    def gameOver(self, winningTeam):
        self.gameInterface.gameOver(winningTeam)

    def setPlayer(self, player):
        self.player = player
        self.player.onPrivateChatReceived.addListener(self.privateChat)
        self.reloadGauge.player = player

    def showSettings(self):
        if self.settingsMenu not in self.elements:
            self.elements.append(self.settingsMenu)

    def hideSettings(self):
        if self.settingsMenu in self.elements:
            self.elements.remove(self.settingsMenu)
            self.gameInterface.updateKeyMapping()

    def _setCurrentUpgrade(self, upgradeType):
        self.currentUpgrade = upgradeType
        self.upgradeDisplay.setUpgrade(self.currentUpgrade, self.player)
        self.menuManager.manager.reset()

    def _requestUpgrade(self):
        if self.player.hasTrosball():
            self.sendRequest(ThrowTrosballMsg(self.getTickId()))
        elif self.currentUpgrade is not None:
            self.sendRequest(BuyUpgradeMsg(
                self.currentUpgrade.upgradeType, self.getTickId()))

    def _changeNickname(self):
        if not self.world.canRename():
            self.newMessage('Cannot change nickname after game has started',
                    self.app.theme.colours.errorMessageColour)
            return

        prompt = NicknameBox(self.app)
        @prompt.onClose.addListener
        def _customEntered():
            if prompt.result == DialogResult.OK:
                nickname = prompt.value
                self.sendRequest(ChangeNicknameMsg(self.player.id,
                        nickname.encode()))

        prompt.show()

    def getTickId(self):
        return self.world.lastTickId

    def doAction(self, action):
        '''
        Activated by hotkey or menu.
        action corresponds to the action name in the keyMapping.
        '''
        if action == 'leaderboard':
            self.showPlayerDetails()
            self.menuManager.manager.reset()
        elif action == 'toggle interface':
            self.toggleInterface()
            self.menuManager.manager.reset()
        elif action == 'more actions':
            self.menuManager.showMoreMenu()
        elif action == 'settings':
            self.showSettings()
        elif action == 'maybeLeave':
            self.menuManager.showQuitMenu()
        elif action == 'leave':
            # Disconnect from the server.
            self.gameInterface.disconnect()
        elif action == 'join':
            # Join if currently spectating
            self.gameInterface.spectatorWantsToJoin()
        elif action == 'menu':
            # Return to main menu and show or hide the menu.
            self.menuManager.escape()
        elif action == 'toggle terminal':
            self.gameInterface.toggleTerminal()
        elif self.gameInterface.gameViewer.replay:
            # Replay-specific actions
            if action in ('follow', 'activate upgrade'):
                # Follow the game action
                self.gameInterface.gameViewer.setTarget(None)
        else:
            # All actions after this line should require a player.
            if self.player is None:
                return
            if action == 'respawn':
                self.sendRequest(RespawnRequestMsg(tickId=self.getTickId()))
            elif action == 'ready':
                if self.player.readyToStart or self.player.readyForTournament:
                    self._castGameVote(PlayerIsReadyMsg(self.player.id, False,
                            False))
                else:
                    self._castGameVote(PlayerIsReadyMsg(self.player.id, True,
                            False))
            elif action in self.upgradeMap:
                self._setCurrentUpgrade(self.upgradeMap[action])
            elif action == 'no upgrade':
                self._setCurrentUpgrade(None)
                self.menuManager.manager.reset()
            elif action == 'abandon':
                self.abandon(self.player)
                self.menuManager.manager.reset()
            elif action == 'chat':
                self.chat()
                self.menuManager.manager.reset()
            elif action == 'select upgrade':
                self.menuManager.showBuyMenu()
            elif action == 'activate upgrade':
                self._requestUpgrade()
            elif action == 'change nickname':
                self._changeNickname()
            elif action not in ('jump', 'right', 'left', 'down'):
                log.warning('Unknown action: %r', action)

    def newMessage(self, text, colour=None, error=False):
        if colour is None:
            if error:
                colour = self.app.theme.colours.errorMessageColour
            else:
                colour = self.app.theme.colours.grey
        self.currentMessages.newMessage(text, colour)

    def privateChat(self, text, sender):
        # Destined for the one player
        text = " (private): " + text
        self.newChat(text, sender)

    def newChat(self, text, sender):
        if sender is None:
            self.chatBox.newServerMessage(text)
        else:
            colour = self.app.theme.colours.chatColour(sender.team)
            self.chatBox.newMessage(text, sender.nick, colour)

    def endInput(self):
        if self.input:
            self.elements.remove(self.input)
            self.input = None
        if self.inputText:
            self.elements.remove(self.inputText)
            self.inputText = None
        if self.unobtrusiveGetter:
            self.elements.remove(self.unobtrusiveGetter)
            self.unobtrusiveGetter = None
        if self.menuManager not in self.elements:
            self.elements.append(self.menuManager)
        self.input = self.inputText = None


    def inputStarted(self):
        self.elements.append(self.input)
        self.elements.append(self.inputText)
        self.input.onEsc.addListener(lambda sender: self.endInput())
        self.input.onEnter.addListener(lambda sender: self.endInput())

        try:
            self.elements.remove(self.menuManager)
        except ValueError:
            pass

    def chat(self):
        if not self.player:
            return

        if self.chatBox.isOpen():
            self.chatBox.close()
        else:
            pygame.key.set_repeat(300, 30)
            self.chatBox.open()
            self.chatBox.setPlayer(self.player)

    def abandon(self, player):
        '''
        Called when a player says they wish to abandon their upgrade.
        '''
        if player.upgrade:
            addOn = 'upgrade'
            if type(player.upgrade) == Grenade:
                self.newMessage('Cannot abandon an active grenade!',
                        self.app.theme.colours.errorMessageColour)
                return
        else:
            return

        message = 'Really abandon your ' + addOn + ' (Y/N)'
        self.endInput()
        self.unobtrusiveGetter = YesNoGetter(self.app, Location(
                FullScreenAttachedPoint((0,0), 'center'), 'center'), message,
                self.app.screenManager.fonts.unobtrusivePromptFont,
                self.app.theme.colours.unobtrusivePromptColour, 3)
        self.elements.append(self.unobtrusiveGetter)

        def gotValue(abandon):
            if self.unobtrusiveGetter:
                self.elements.remove(self.unobtrusiveGetter)
                self.unobtrusiveGetter = None
            if abandon:
                self.sendRequest(
                    AbandonUpgradeMsg(player.id, tickId=self.getTickId()))

        self.unobtrusiveGetter.onGotValue.addListener(gotValue)

    def upgradeUsed(self, player):
        upgrade = player.upgrade
        if upgrade is None:
            upgrade = 'an upgrade'
        else:
            upgrade = str(upgrade)

        if type(player.upgrade) == Grenade:
            message = '%s has thrown a %s' % (player.nick, player.upgrade)
        elif type(player.upgrade) in [Ninja, Turret]:
            message = '%s has become a %s' % (player.nick, player.upgrade)
        elif type(player.upgrade) == Directatorship:
            message = '%s has become a Directator' % player.nick
        elif type(player.upgrade) in [Ricochet, Shoxwave]:
            message = '%s is using %s' % (player.nick, player.upgrade)
        else:
            message = '%s is using a %s' % (player.nick, player.upgrade)

        self.newMessage(message)

    def showPlayerDetails(self):
        self.gameInterface.gameViewer.toggleLeaderBoard()

    def toggleInterface(self):
        self.gameInterface.gameViewer.toggleInterface()

class UpgradeDisplay(framework.CompoundElement):

    def __init__(self, app):
        super(UpgradeDisplay, self).__init__(app)
        self.player = None
        self.upgrade = None

    def refresh(self):
        self.setUpgrade(self.upgrade, self.player)

    def setUpgrade(self, upgradeType, player):
        self.player = player
        self.upgrade = upgradeType
        if player is None or upgradeType is None:
            self.elements = []
        else:
            pos = Location(Screen(0.6, 0), 'midtop')
            image = self.app.theme.sprites.upgradeImage(upgradeType)
            area = Area(
                RelativePoint(Screen(0.6, 0), (0, 52)),
                ScaledSize(50, 10), 'midtop')
            self.elements = [
                PictureElement(self.app, image, pos),
            ]

            if upgradeType.enabled:
                self.elements.append(
                    StarGauge(self.app, area, player, upgradeType))
            else:
                self.elements.append(
                    TextElement(self.app, 'DISABLED',
                        self.app.screenManager.fonts.ingameMenuFont,
                        Location(CanvasX(620, 68), 'midbottom'),
                        self.app.theme.colours.errorMessageColour))

class TrajectoryOverlay(framework.Element):
    def __init__(self, app, viewManager, upgradeDisplay):
        super(TrajectoryOverlay, self).__init__(app)
        self.viewManager = viewManager
        self.upgradeDisplay = upgradeDisplay
        self.enabled = False

    def setEnabled(self, enable):
        self.enabled = enable

    def isActive(self):
        if not self.enabled:
            return False

        player = self.viewManager.getTargetPlayer()

        if not player.hasTrosball() and (self.upgradeDisplay.upgrade is None or self.upgradeDisplay.upgrade.name != 'Grenade' or player.getTeamStars() < Grenade.requiredStars):
            # Nothing to throw yet
            return False

        if player is None:
            return False

        trajectory = player.currentProjectileTrajectory
        if trajectory is None:
            return False
        return True

    def draw(self, surface):
        if not self.isActive():
            return

        player = self.viewManager.getTargetPlayer()
        trajectory = player.currentProjectileTrajectory

        focus = self.viewManager._focus
        area = self.viewManager.sRect

        points = list(trajectory.predictedTrajectoryPoints())
        numPoints = len(points)
        greenToYellow = [(i, 255, 0) for i in range(0, 255, 255 / (numPoints / 2 + numPoints % 2))]
        yellowToRed = [(255, i, 0) for i in range(255, 0, -255 / (numPoints / 2))]

        colours = [(0, 255, 0)] + greenToYellow + yellowToRed

        lastPoint = None
        for item in zip(points, colours):
            point = item[0]
            colour = item[1]
            # Set first point
            if lastPoint is None:
                lastPoint = point
                adjustedPoint = mapPosToScreen(point, focus, area)
                pygame.draw.circle(surface, colour, adjustedPoint, 2)
                continue


            adjustedPoint = mapPosToScreen(point, focus, area)
            adjustedLastPoint = mapPosToScreen(lastPoint, focus, area)

            pygame.draw.line(surface, colour, adjustedLastPoint, adjustedPoint, 5)

            lastPoint = point

        pygame.draw.circle(surface, colour, adjustedPoint, 2)

        radius = int(trajectory.explosionRadius() * MAP_TO_SCREEN_SCALE)
        if radius > 0:
            pygame.draw.circle(surface, (0,0,0), adjustedPoint, radius, 2)



import datetime
from functools import partial
import logging
import os
import pickle
import random

from django.contrib.auth import authenticate
from django.core import management

from trosnoth import data
from trosnoth.data import getPath, makeDirs
from trosnoth import dbqueue
from trosnoth.djangoapp.models import (
    User, TrosnothUser, AchievementProgress, TrosnothServerSettings,
)
from trosnoth.game import LocalGame
from trosnoth.gamerecording.stats import ServerGameStats
from trosnoth import rsa
from trosnoth.manholehelper import AuthServerManholeHelper
from trosnoth.model.mapLayout import LayoutDatabase
from trosnoth import murmur
from trosnoth.network import authcommands
from trosnoth.network.manhole import Manhole
from trosnoth.network.server import TrosnothServerFactory
from trosnoth.network.networkDefines import serverVersion
from trosnoth.settings import (
    AuthServerSettings, TournamentSettings)
from trosnoth.utils.event import Event
from trosnoth.utils.utils import hasher, timeNow, initLogging

from trosnoth.gamerecording.achievementlist import availableAchievements

from twisted.protocols import amp
from twisted.internet import reactor, defer
from twisted.internet.protocol import Factory, ClientCreator, Protocol
from twisted.internet.error import CannotListenError, ConnectError
from twisted.cred import portal, checkers
from twisted.conch import manhole_ssh

log = logging.getLogger('authserver')

MAX_GAMES = 1
GAME_KIND = 'Trosnoth1'

GAME_VERIFY_TIME = 60
REPLAY_SUB_PATH = os.path.join('authserver', 'public', 'trosnoth', 'replays')


class AuthenticationProtocol(amp.AMP):
    '''
    Trosnoth authentication server which is used when running a game server
    which keeps track of users.
    @param dataPath: path to the location for this authentication server to
        store its data. If not specified, $HOME/.trosnoth/authserver will be
        used. If the path does not exist, it will be created.
    '''

    def connectionMade(self):
        super(AuthenticationProtocol, self).connectionMade()
        self.user = None
        self.token = None
        log.info('New connection.')

    def connectionLost(self, reason):
        log.info('Connection lost.')

    @authcommands.GetPublicKey.responder
    def getPublicKey(self):
        return {
            'e': self.factory.pubKey['e'],
            'n': self.factory.pubKey['n'],
        }

    @authcommands.ListGames.responder
    def listGames(self):
        games = []
        for gameId in self.factory.servers.iterkeys():
            games.append({
                'id': gameId,
                'game': GAME_KIND,
                'version': serverVersion,
            })

        return {'games': games}

    @authcommands.ListOtherGames.responder
    def listOtherGames(self):
        d = self.factory.getRegisteredGames()

        @d.addCallback
        def gotGames(rGames):
            result = []
            for rGame in rGames:
                result.append({
                    'game': rGame.game,
                    'version': rGame.version,
                    'ip': rGame.host,
                    'port': rGame.port,
                })
            return {
                'games': result,
            }
        return d

    @authcommands.RegisterGame.responder
    def registerGame(self, game, version, port):
        host = self.transport.getPeer().host
        self.factory.registerGame(game, version, host, port)
        return {}

    @authcommands.CreateGame.responder
    def createGame(self, game):
        if game != GAME_KIND:
            raise authcommands.CannotCreateGame()
        if len(self.factory.servers) >= MAX_GAMES:
            raise authcommands.CannotCreateGame()
        return {
            'id': self.factory.createGame(),
            'version': serverVersion,
        }

    @authcommands.ConnectToGame.responder
    def connectToGame(self, id):
        if self.user is None:
            raise authcommands.NotAuthenticated()
        try:
            server = self.factory.servers[id]
        except KeyError:
            raise authcommands.GameDoesNotExist()

        tag = server.authTagManager.newTag(self.user)
        nick = self.user.getNick()

        self.factory.broadcast('%s has joined the game' % (nick,))
        return {
            'port': server.getTCPPort(),
            'authTag': tag,
            'nick': nick,
        }

    @authcommands.GetSupportedSettings.responder
    def getSupportedSettings(self):
        settings = ['password']
        return {
            'result': settings,
        }

    @authcommands.SetPassword.responder
    def setUserPassword(self, password):
        if self.user is None:
            raise authcommands.NotAuthenticated()
        password = self._decodePassword(password)
        self.user.setPassword(password)
        return {}

    @authcommands.GetAuthToken.responder
    def getAuthToken(self):
        self.token = ''.join(str(random.randrange(256)) for i in xrange(16))
        return {
            'token': self.token
        }

    @authcommands.PasswordAuthenticate.responder
    def passwordAuthenticate(self, username, password):
        username = username.lower()
        password = self._decodePassword(password)
        if password is None:
            return {'result': False}    # Bad auth token used.

        d = self.factory.authManager.authenticateUser(username, password)

        @d.addCallback
        def authSucceeded(user):
            self.user = user
            return {'result': True}

        @d.addErrback
        def authFailed(failure):
            return {'result': False}

        return d

    @authcommands.CreateUserWithPassword.responder
    def createUserWithPassword(self, username, password):
        if self.factory.settings.allowNewUsers:
            nick = username
            username = username.lower()
            password = self._decodePassword(password)
            if password is None:
                return {'result': 'Authentication token failure.'}

            authMan = self.factory.authManager
            if authMan.checkUsername(username):
                return {'result': 'That username is taken.'}
            self.user = authMan.createUser(username, password, nick)
        else:
            return {'result': self.factory.settings.privateMsg}
        return {'result': ''}

    def _decodePassword(self, password):
        if self.token is None:
            return None
        token, self.token = self.token, None
        passwordData = rsa.decrypt(password, self.factory.privKey)
        if passwordData[:len(token)] != token:
            return None
        return passwordData[len(token):]

SALT = 'Trosnoth'


class AuthManager(object):
    '''
    Manages user accounts on the system.
    '''

    def __init__(self, dataPath):
        self.dataPath = dataPath
        self.tags = {}      # auth tag -> user id
        self.settings = AuthServerSettings(dataPath)
        self.tournamentSettings = TournamentSettings(dataPath)

    def checkUsername(self, username):
        '''
        Returns True or False, depending on whether the given username is
        already in use.
        '''
        try:
            TrosnothUser.fromUser(username=username)
        except User.DoesNotExist:
            return False
        return True

    def newTagManager(self):
        '''
        Returns an authentication tag manager for this set of users.
        '''
        return AuthTagManager(self)

    def authenticateUser(self, username, password):
        '''
        If a username exists with the given password, returns the user,
        otherwise returns None.
        '''
        hash1 = hasher(SALT + password).digest()
        username = username.lower()
        try:
            trosnothUser = TrosnothUser.fromUser(username=username)
        except User.DoesNotExist:
            return defer.fail()

        if not trosnothUser.oldPasswordHash:
            # Just use Django auth
            djangoUser = authenticate(username=username, password=password)
            if djangoUser is None:
                return defer.fail(ValueError('Authentication failed'))
            if not djangoUser.is_active:
                return defer.fail(ValueError('User deactivated'))
            user = AuthenticatedUser(self, username)
        else:
            # Old Trosnoth auth, only exists for backward compatibility
            hash2 = bytes(trosnothUser.oldPasswordHash)
            if hash1 != hash2:
                return defer.fail(ValueError('Incorrect password'))

            # Put the password into Django
            trosnothUser.user.set_password(password)
            trosnothUser.user.save()
            trosnothUser.oldPasswordHash = ''
            trosnothUser.save()

            user = AuthenticatedUser(self, username)

        user.seen()
        return defer.succeed(user)

    def getUserByName(self, username):
        return AuthenticatedUser(self, username.lower())

    def createUser(self, username, password, nick=None):
        username = username.lower()
        if self.checkUsername(username):
            raise ValueError('user %r already exists' % (username,))
        User.objects.create_user(username, password=password)

        user = AuthenticatedUser(self, username)
        user.setPassword(password)
        user.seen()
        if nick is not None:
            user.setNick(nick)
        return user

    def getNick(self, username):
        return TrosnothUser.fromUser(username=username).nick


class AuthenticationFactory(Factory):
    protocol = AuthenticationProtocol
    authManagerClass = AuthManager

    def __init__(self, dataPath=None, gameType=None):
        if dataPath is None:
            dataPath = getPath(data.user, 'authserver')
        if gameType is None:
            gameType = 'normal'
        makeDirs(dataPath)
        self.dataPath = dataPath

        self.onPrimaryGameChanged = Event()
        self.primaryGameId = None

        self.authManager = self.authManagerClass(dataPath)
        self.gameType = gameType
        self.pubKey, self.privKey = self.loadKeys()
        self.servers = {}     # Game id -> game.
        self.nextId = 0
        self.registeredGames = []
        self.gameStats = {}

        self.layoutDatabase = LayoutDatabase()

        self.notifier = self.settings.createNotificationClient()
        if self.notifier is not None:
            self.notifier.startService()

    @property
    def settings(self):
        return self.authManager.settings

    def getPlayerCount(self):
        '''
        Returns the total number of online players in all games.
        '''
        result = 0
        for server in self.servers.itervalues():
            for p in server.game.world.players:
                if not p.bot:
                    result += 1
        return result

    def loadKeys(self):
        '''
        Loads public and private keys from disk or creates them and saves them.
        '''
        keyPath = os.path.join(self.dataPath, 'keys')
        try:
            pub, priv = pickle.load(open(keyPath, 'rb'))
        except IOError:
            pub, priv = rsa.newkeys(self.settings.keyLength)
            pickle.dump((pub, priv), open(keyPath, 'wb'), 2)

        return pub, priv

    def createGame(self):
        '''
        Creates a new game and returns the game id.
        '''
        gameId = self.nextId
        self.nextId += 1
        authTagMan = self.authManager.newTagManager()
        halfWidth, height = self.settings.lobbySize

        game = LocalGame(
            layoutDatabase=self.layoutDatabase,
            halfMapWidth=halfWidth,
            mapHeight=height,
            maxPerTeam=self.settings.maxPerTeam,
            authTagManager=authTagMan,
            saveReplay=True,
            gamePrefix='Server',
            replayPath=REPLAY_SUB_PATH,
        )
        try:
            game.world.setNextGameType(self.gameType)
        except ValueError:
            log.warning('Unknown game type: %s' % (self.gameType,))
            self.gameType = 'normal'
            game.world.setNextGameType(self.gameType)
        server = TrosnothServerFactory(game, authTagMan)

        def onShutdown():
            server.stopListening()
            del self.servers[gameId]
            del self.gameStats[game]
            game.world.onGameStateChanged.removeListener(gameStateChanged)
            if gameId == self.primaryGameId:
                self.primaryGameId = None
                self.onPrimaryGameChanged(None)
        server.onShutdown.addListener(onShutdown)
        server.startListening()

        self.gameStats[game] = None
        gameStateChanged = partial(self.gameStateChanged, game=game)
        game.world.onGameStateChanged.addListener(gameStateChanged)

        if murmur.state == 'initialised':
            game.world.onChangeVoiceChatRooms.addListener(murmur.setupRooms)

        log.info('Created game (id=%d)' % (gameId,))
        self.servers[gameId] = server
        self.broadcast('a new game has been created')
        if self.primaryGameId is None:
            self.primaryGameId = gameId
            self.onPrimaryGameChanged(server)

        return gameId

    def gameStateChanged(self, game):
        countStats = game.world.state.statsCount

        gameStats = self.gameStats[game]
        if gameStats and not countStats:
            # Stopping game
            gameStats.stopAndSave()
            self.gameStats[game] = None

        elif countStats and not gameStats:
            # Starting game
            self.gameStats[game] = ServerGameStats(game)

    def broadcast(self, message):
        if self.notifier is None:
            return
        try:
            self.notifier.broadcast('%s: %s' % (
                self.settings.hostName, message))
        except:
            log.error('Error broadcasting message', exc_info=True)

    @defer.inlineCallbacks
    def registerGame(self, game, version, host, port):
        '''
        Registers a remote game with this server.
        '''
        settings = TrosnothServerSettings.get()
        if not settings.allowRemoteGameRegistration:
            return

        rGame = RegisteredGame(game, version, host, port)
        result = yield rGame.verify()
        if result:
            log.info('Registered game on %s:%s', host, port)
            self.registeredGames.append(rGame)
        else:
            log.info('Failed to connect to game on %s:%s', host, port)
            raise authcommands.PortUnreachable()

    def getRegisteredGames(self):
        '''
        Returns a list of registered games which are running.
        '''
        if len(self.registeredGames) == 0:
            return defer.succeed([])

        result = []
        d = defer.Deferred()
        remaining = [len(self.registeredGames)]

        def gameVerified(success, rGame):
            if success:
                result.append(rGame)
            else:
                self.registeredGames.remove(rGame)

            remaining[0] -= 1
            if remaining[0] == 0:
                d.callback(result)

        for rGame in self.registeredGames:
            rGame.verify().addCallback(gameVerified, rGame)

        return d


class RegisteredGame(object):
    '''
    Represents a remote game that has been registered with this server.
    '''

    def __init__(self, game, version, host, port):
        self.game = game
        self.version = version
        self.host = host
        self.port = port
        self.lastVerified = 0
        self.running = True

    @defer.inlineCallbacks
    def verify(self):
        '''
        If more than GAME_VERIFY_TIME has passed since the last verification,
        connects to the game to check whether it is still running. Returns a
        deferred whose callback will be executed with True or False depending
        on whether the game is still running.
        '''
        t = timeNow()
        if t - self.lastVerified < GAME_VERIFY_TIME:
            defer.returnValue(self.running)
            return
        self.lastVerified = t

        try:
            yield ClientCreator(reactor, Protocol).connectTCP(
                self.host, self.port, timeout=5)
        except ConnectError:
            self.running = False
            defer.returnValue(False)
            return

        self.running = True
        defer.returnValue(True)


class AuthTagManager(object):
    '''
    Provides a gateway for the Validator to confirm that a player has
    authenticated and to check the player's userid.
    '''

    def __init__(self, authManager):
        self.authManager = authManager
        self.tags = {}      # auth tag -> user id

    def checkUsername(self, username):
        '''
        Returns True or False, depending on whether the given username is
        already in use.
        '''
        return self.authManager.checkUsername(username)

    def newTag(self, user):
        tag = random.randrange(1 << 64)
        self.tags[tag] = user
        return tag

    def checkAuthTag(self, tag):
        '''
        Returns the authenticated user corresponding to the auth tag, or
        None if the tag is not recognised.
        Don't forget to call discardAuthTag() when the authentication tag is no
        longer valid.
        '''
        return self.tags.get(tag, None)

    def discardAuthTag(self, tag):
        try:
            del self.tags[tag]
        except KeyError:
            pass


class AuthenticatedUser(object):
    '''
    Represents a user which has been authenticated on the system.
    '''
    achievementDefs = availableAchievements

    def __init__(self, authManager, username):
        self.authManager = authManager
        self.username = username = username.lower()

    def __eq__(self, other):
        if (isinstance(other, AuthenticatedUser) and other.username ==
                self.username):
            return True
        return False

    def __hash__(self):
        return hash(self.username)

    def isElephantOwner(self):
        return self.username in self.authManager.settings.elephantOwners

    def getNick(self):
        return TrosnothUser.fromUser(username=self.username).nick

    def getDefaultTeam(self):
        # Not yet implemented. Could be used to set what tournament team this
        # player usually playes on.
        return ''

    def setNick(self, nick):
        @dbqueue.add
        def writeNickToDB():
            user = TrosnothUser.fromUser(username=self.username)
            if nick != user.nick:
                user.nick = nick
                user.save()

    def setPassword(self, password):
        # Don't put DB write in a queue as user will expect it to take place
        # immediately.
        user = User.objects.get(username=self.username)
        user.set_password(password)
        user.save()
        trosnothUser = TrosnothUser.fromUser(pk=user.pk)
        trosnothUser.oldPasswordHash = ''
        trosnothUser.save()

    def getAchievementRecord(self, achievementId):
        user = TrosnothUser.fromUser(username=self.username)
        try:
            achievement = AchievementProgress.objects.get(
                user=user, achievementId=achievementId)
        except AchievementProgress.DoesNotExist:
            achievement = AchievementProgress(
                user=user, achievementId=achievementId)
            achievement.save()
        return achievement

    def achievementUnlocked(self, achievementId):
        @dbqueue.add
        def writeUnlockedAchievementToDB():
            user = TrosnothUser.fromUser(username=self.username)
            try:
                achievement = AchievementProgress.objects.get(
                    user=user, achievementId=achievementId)
            except AchievementProgress.DoesNotExist:
                achievement = AchievementProgress(
                    user=user, achievementId=achievementId)
            if not achievement.unlocked:
                achievement.unlocked = True
                achievement.save()

    def seen(self):
        now = datetime.datetime.now()
        @dbqueue.add
        def writeSeenTimeToDB():
            user = TrosnothUser.fromUser(username=self.username)
            user.lastSeen = now
            user.save()


def startServer(
        port=6787, dataPath=None, manholePort=6799, password=None,
        webPort=None, gameType=None):

    # Ensure that any database migrations have happened
    management.call_command('migrate')

    # If murmur communication is enabled, try to connect
    if murmur.init() == 'initialised':
        reactor.addSystemEventTrigger(
            'before', 'shutdown', murmur.tearDownRooms)

    dbqueue.init()

    pf = AuthenticationFactory(dataPath, gameType)

    def getManholeFactory(namespace, password):
        realm = manhole_ssh.TerminalRealm()

        # If we don't do this, the server will generate an exception when
        # you resize the SSH window
        def windowChanged(self, size):
            pass
        realm.sessionFactory.windowChanged = windowChanged

        def getManhole(_):
            return Manhole(namespace)
        realm.chainedProtocolFactory.protocolFactory = getManhole
        p = portal.Portal(realm)

        # Username/Password authentication
        passwordDB = checkers.InMemoryUsernamePasswordDatabaseDontUse()
        passwordDB.addUser('trosnoth', password)
        p.registerChecker(passwordDB)

        factory = manhole_ssh.ConchFactory(p)
        return factory

    if password is None:
        password = ''.join(random.choice('0123456789') for i in xrange(6))

    namespace = {}
    namespace['authFactory'] = pf
    namespace['helper'] = AuthServerManholeHelper(pf)
    factory = getManholeFactory(namespace, password)

    try:
        reactor.listenTCP(manholePort, factory)
    except CannotListenError:
        log.error('Error starting manhole on port %d', manholePort)
    else:
        log.warning(
            'SSH manhole started on port %d with password %r',
            manholePort, password)

    if webPort is not None:
        from trosnoth.web.server import startWebServer
        startWebServer(pf, webPort)

    try:
        reactor.listenTCP(port, pf)
    except CannotListenError:
        log.error('Error listening on port %d.', port)
    else:
        log.info(
            'Started Trosnoth authentication server on port %d.', port)
        reactor.run()


def main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option(
        '-p', '--port', action='store', dest='port', default=6787,
        help='which port to run the authentication server on')
    parser.add_option(
        '-D', '--datapath', action='store', dest='dataPath', default=None,
        help='where to store the authentication server data')
    parser.add_option(
        '-m', '--manhole', action='store', dest='manholePort',
        default=6799, help='which port to run the manhole on')
    parser.add_option(
        '--password', action='store', dest='manholePassword', default=None,
        help='the password to use for the manhole')
    parser.add_option(
        '-w', '--webport', action='store', dest='webPort', default='8080',
        help='the port on which to launch the web service. '
        'Default is 8080. To disable web service, use --webport= with no '
        'parameter.')
    parser.add_option(
        '-d', '--debug', action='store_true', dest='debug',
        help='show debug-level messages on console')
    parser.add_option(
        '-l', '--log-file', action='store', dest='logFile',
        help='file to write logs to')
    parser.add_option(
        '-t', '--gametype', action='store', dest='gameType', default=None,
        help='the game type to run')
    parser.add_option(
        '--profile', action='store_true', dest='profile',
        help='dump kcachegrind profiling data to trosnoth.log')

    options, args = parser.parse_args()
    if len(args) > 0:
        parser.error('no arguments expected')

    initLogging(options.debug, options.logFile)

    if not options.webPort:
        webPort = None
    else:
        try:
            webPort = int(options.webPort)
        except ValueError:
            options.error('Invalid port: %r' % (options.webPort,))

    kwargs = dict(
        port=int(options.port), dataPath=options.dataPath,
        manholePort=int(options.manholePort),
        password=options.manholePassword, webPort=webPort,
        gameType=options.gameType,
    )

    if options.profile:
        runWithProfiling(**kwargs)
    else:
        startServer(**kwargs)


def runWithProfiling(**kwargs):
    import cProfile
    from trosnoth.utils.profiling import KCacheGrindOutputter
    prof = cProfile.Profile()

    try:
        prof.runcall(startServer, **kwargs)
    except SystemExit:
        pass
    finally:
        kg = KCacheGrindOutputter(prof)
        with open('server.log', 'wb') as f:
            kg.output(f)



if __name__ == '__main__':
    main()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

def wrapWords(app, text, font, width, breakWords = True):
    '''
    Takes a section of text, a font with which it will be rendered, and the
    maximum size that a line can be, and returns a list of the new strings.
    @param text The text to split
    @param font The font that the text will be rendered with
    @param width The maximum width (in pixels) of a single line of text
    @param breakWords If this is set to true, words will be allowed to be
            broken in half if it may look neater (annotated with a hyphen).
            If not, words will only break if absolutely necessary.
    '''
    lines = text.split('\n')
    # Determine if the line needs word-wrapping:
    x = 0
    while x < len(lines):
        line = lines[x]
        lastSp = -1
        p = 0
        while p < len(line):
            if line[p] == " ":
                lastSp = p

            p += 1
            if font.size(app, line[:p])[0] > width:
                # we need to wordwrap
                if lastSp == -1 or (p - lastSp > 20 and breakWords):
                    spFound = False
                    for t in range(p, p + 5):
                        if t >= len(line) or line[t] == ' ':
                            spFound = True
                            break
                    if spFound and breakWords:
                        # But there's a space coming up within the next 5
                        # characters - let's even things up a bit by giving
                        # it a few more from this end.
                        lines[x] = line[:p - t - 2] + "-"
                        lines.insert(x+1, line[p-t-2:])
                        line = line[:p - t - 2] + "-"
                    else:
                        lines[x] = line[:p] + "-"
                        lines.insert(x+1, line[p:])
                        line = line[:p] + "-"
                else:
                    lines[x] = line[:lastSp]
                    lines.insert(x + 1, line[lastSp + 1:])
                    line = lines[x]
                break
        x += 1
    return lines



import pygame

defaultAnchor = 'topleft'


def uniqueColour(colours):
    '''Returns a greyish colour that is not in the given list of colour
    triples. Useful for colourkeying.'''
    result = [0, 0, 0]
    i = 0
    while True:
        result[i] += 1
        rTuple = tuple(result)
        for colour in colours:
            if colour == rTuple:
                break
        else:
            return rTuple
        i = (i + 1) % 3


def addPositions(p1, p2):
    return tuple([p1[i] + p2[i] for i in 0, 1])


def canvasIntervalToScreen(app, val):
    return max(1, app.screenManager.scaleFactor * val)


class Scalar(object):
    def __init__(self, val):
        self.val = val

    def getVal(self, app):
        return self.val

    def __repr__(self):
        return "Scalar(%r)" % (self.val,)


class ScaledScalar(object):
    def __init__(self, val):
        self.val = val
        self._val = None
        self._scaleFactor = None

    def getVal(self, app):
        sf = app.screenManager.scaleFactor
        if self._scaleFactor != sf:
            self._val = canvasIntervalToScreen(app, self.val)
            self._scaleFactor = sf
        return self._val

    def __repr__(self):
        return "ScaledScalar (%d)" % (self.val,)


class FullScreenAttachedPoint(object):
    def __init__(self, distance, attachedAt='topleft'):
        self.val = distance
        self.attachedAt = attachedAt

    def getPoint(self, app):
        if hasattr(self.val, 'getSize'):
            val = self.val.getSize(app)
        else:
            val = self.val
        pos = getattr(app.screenManager.rect, self.attachedAt)
        return addPositions(pos, val)


class ScaledScreenAttachedPoint(object):
    def __init__(self, distance, attachedAt='topleft'):
        self.val = distance
        self.attachedAt = attachedAt

    def getPoint(self, app):
        if hasattr(self.val, 'getSize'):
            val = self.val.getSize(app)
        else:
            val = self.val
        pos = getattr(app.screenManager.scaledRect, self.attachedAt)
        return addPositions(pos, val)


class AttachedPoint(object):
    '''
    @param attachedTo_Rect: should be a reference to a function which
            returns a pygame.Rect
    '''

    def __init__(self, val, attachedTo_Rect, attachedAt='topleft'):
        self.val = val
        self.attachedAt = attachedAt
        self.attachedTo_Rect = attachedTo_Rect

    def getPoint(self, app):
        pos = getattr(self.attachedTo_Rect(), self.attachedAt)
        if hasattr(self.val, 'getSize'):
            val = self.val.getSize(app)
        else:
            val = self.val

        return addPositions(pos, val)


class RelativePoint(object):
    '''
    A point relative to another point.
    '''

    def __init__(self, point, size):
        self.size = size
        self.point = point

    def __repr__(self):
        return 'RelativePoint: (%s plus %s)' % (
            repr(self.point), repr(self.size))

    def getPoint(self, app):
        pos = self.point.getPoint(app)
        if hasattr(self.size, 'getSize'):
            diff = self.size.getSize(app)
        else:
            diff = self.size
        return addPositions(pos, diff)


class FullScreenSize(object):
    def getSize(self, app):
        return app.screenManager.size

    def __repr__(self):
        return "Full Screen Size"


class ScaledScreenSize(object):
    def getSize(self, app):
        return app.screenManager.scaledSize

    def __repr__(self):
        return "Scaled Screen Size"


# An image loaded from file
class NewImage(object):
    def __init__(self, nameOrImage, colourkey=True):
        if isinstance(nameOrImage, str):
            self.name = nameOrImage
            self.image = None
        else:
            assert isinstance(nameOrImage, pygame.Surface)
            self.image = nameOrImage
        self.colourkey = colourkey

    def getImage(self, app):
        if self.image is not None:
            return self.image
        self.image = pygame.image.load(self.name).convert()
        if self.colourkey:
            self.image.set_colorkey(self.image.get_at((0, 0)))
        return self.image

    def getSize(self, app):
        return self.getImage(app).get_size()


class SizedImage(NewImage):
    def __init__(self, nameOrImage, size, colourkey=True):
        super(SizedImage, self).__init__(nameOrImage, colourkey)
        self.sizedImage = None
        # Keep a record of whether the size of the screen changes
        self.lastSize = None
        self.size = size

    def getImage(self, app):
        if (self.sizedImage is not None and self.lastSize ==
                self._getCurrentSize(app)):
            return self.sizedImage
        img = super(SizedImage, self).getImage(app)
        self.lastSize = self._getCurrentSize(app)
        self.sizedImage = pygame.transform.smoothscale(img, self.lastSize)
        return self.sizedImage

    def _getCurrentSize(self, app):
        if hasattr(self.size, 'getSize'):
            return self.size.getSize(app)
        else:
            return self.size


class TextImage(NewImage):
    '''Shows text at a specified screen location.'''

    def __init__(
            self, text, font, colour=(0, 128, 0), bgColour=None,
            antialias=True):
        self.colour = colour
        self.bgColour = bgColour
        self._text = str(text)
        self.font = font
        self.antialias = antialias
        self.image = None

    def setColour(self, colour):
        self.colour = colour
        self.image = None

    def setBackColour(self, colour):
        self.bgColour = colour
        self.image = None

    def setFont(self, font):
        self.font = font
        self.image = None

    def setText(self, text):
        self._text = str(text)
        self.image = None

    def refresh(self):
        self.image = None

    def getText(self):
        return self._text
    text = property(getText, setText)

    def getImage(self, app):
        if not self.image:
            self.image = self.font.render(
                app, self._text, self.antialias, self.colour, self.bgColour)
        return self.image


class Location(object):
    '''
    @param  point   the point that this location is attached to
    @param  anchor  a string representing the rectangle attribute which this
                    location will reposition
    '''

    def __init__(self, point, anchor=defaultAnchor):
        if anchor == 'centre':
            anchor = 'center'
        self.point = point
        self.anchor = anchor

    def __repr__(self):
        return "Location: %s anchored at %s" % (repr(self.point), self.anchor)

    def apply(self, app, rect):
        '''Moves the rect so that it is anchored according to this anchor.'''
        point = self.point
        if hasattr(point, 'getPoint'):
            point = point.getPoint(app)
        setattr(rect, self.anchor, point)


# Location relative to another location
# @param size Can be a tuple or a ScaledPoint
class RelativeLocation(Location):
    def __init__(self, location, size):
        self.size = size
        self.location = location

    def __repr__(self):
        return 'RelativeLocation: (%s plus %s)' % (
            repr(self.location), repr(self.size))

    def apply(self, app, rect):
        self.location.apply(app, rect)
        if hasattr(self.size, 'getSize'):
            diff = self.size.getSize(app)
        else:
            diff = self.size
        rect.topleft = addPositions(diff, rect.topleft)


class Area(Location):
    '''
    @param  point   the point that this area is attached to
    @param  size    the size of this area
    @param  anchor  a string representing which of the rectangle's attributes
                    the point will be attached to
    '''

    def __init__(self, point, size, anchor=defaultAnchor):
        Location.__init__(self, point, anchor)
        self.size = size

    def apply(self, app, rect):
        '''Moves and resizes this rect so that it covers this area.'''
        if hasattr(self.size, 'getSize'):
            rect.size = self.size.getSize(app)
        else:
            rect.size = self.size
        Location.apply(self, app, rect)

    def getRect(self, app):
        result = pygame.Rect(0, 0, 0, 0)
        self.apply(app, result)
        return result

    def __repr__(self):
        return "Area: point is %s, size is %s, anchored at %s" % (
            repr(self.point), repr(self.size), self.anchor)


class PygameEvent(object):
    def __init__(self, **kwargs):
        self._repr = 'PygameEvent(%s)' % (', '.join(
            '%s=%s' % a for a in kwargs.iteritems()))
        for k, v in kwargs.iteritems():
            setattr(self, k, v)

    def __repr__(self):
        return self._repr


def translateEvent(event, amount):
    result = PygameEvent(type=event.type, pos=(
        event.pos[0] - amount[0], event.pos[1] - amount[1]))
    for name in ['button', 'buttons']:
        try:
            v = getattr(event, name)
        except AttributeError:
            pass
        else:
            setattr(result, name, v)
    return result


# For ease of use
def ScaledLocation(x, y, anchor=defaultAnchor):     # pragma: noqa
    return Location(ScaledPoint(x, y), anchor)


def ScaledArea(x, y, width, height, anchor=defaultAnchor):      # pragma: noqa
    return Area(ScaledPoint(x, y), ScaledSize(width, height), anchor)


class Region(object):
    '''
    Defines a region using any of various anchors. The values passed in should
    be one of:
        * Abs(a, b) - absolute coordinates.
        * Canvas(a, b) - takes (a, b) on a 1024x768 surface and maps to the
            screen.
        * Screen(a, b) - takes floats a and b in the range of 0..1 and scales
            them to the screen size.
    '''
    STANDARD_ARGS = {
        'topleft':      ('0', '0'),
        'topright':     ('2', '0'),
        'bottomleft':   ('0', '2'),
        'bottomright':  ('2', '2'),
        'size':         ('s', 's'),
        'midtop':       ('1', '0'),
        'midbottom':    ('1', '2'),
        'midleft':      ('0', '1'),
        'midright':     ('2', '1'),
        'centre':       ('1', '1'),
        'center':       ('1', '1'),
        'left':         ('0', '-'),
        'right':        ('2', '-'),
        'x':            ('0', '-'),
        'width':        ('s', '-'),
        'w':            ('s', '-'),
        'centrex':      ('1', '-'),
        'centerx':      ('1', '-'),
        'top':          ('-', '0'),
        'bottom':       ('-', '2'),
        'y':            ('-', '0'),
        'height':       ('-', 's'),
        'h':            ('-', 's'),
        'centrey':      ('-', '1'),
        'centery':      ('-', '1'),
    }

    OTHER_ARGS = frozenset(['aspect'])     # The aspect ratio.

    VALID_ARGS = OTHER_ARGS.union(STANDARD_ARGS.keys())

    def __init__(self, **kwargs):
        self.aspect = None
        self.aspect_dir = None

        self._validateArgNames(kwargs)
        self._setRepr(kwargs)

        aspect = None
        if 'aspect' in kwargs:
            aspect = kwargs.pop('aspect')

        xConstraints = {}
        yConstraints = {}
        constraints_2d = []
        constraints_1d = []
        for arg, val in kwargs.iteritems():
            xKind, yKind = self.STANDARD_ARGS[arg]
            if xKind != '-':
                if xKind in xConstraints:
                    raise TypeError(
                        'specifying %s overconstrains region (x)' % (arg,))
                xConstraints[xKind] = arg
            if yKind != '-':
                if yKind in yConstraints:
                    raise TypeError(
                        'specifying %s overconstrains region (y)' % (arg,))
                yConstraints[yKind] = arg

            if xKind == '-':
                c = constraints_1d
                if yKind == 's':
                    method = 'getYVal'
                else:
                    method = 'getYPoint'
            elif yKind == '-':
                c = constraints_1d
                if xKind == 's':
                    method = 'getXVal'
                else:
                    method = 'getXPoint'
            else:
                c = constraints_2d
                if xKind == 's' and yKind == 's':
                    method = 'getSize'
                elif xKind != 's' and yKind != 's':
                    method = 'getPoint'
                else:
                    raise AssertionError(
                        'definition of %r in '
                        'Region.STANDARD_ARGS does not make sense' % (arg,))

            c.append((arg, val, method))

        self.constraints_2d = constraints_2d
        self.constraints_1d = constraints_1d

        xNum = len(xConstraints)
        yNum = len(yConstraints)
        if xNum > 2:
            raise TypeError('Region is overconstrained horizontally')
        if yNum > 2:
            raise TypeError('Region is overconstrained vertically')
        if aspect is None:
            if xNum < 2:
                raise TypeError('Region is underconstrained horizontally')
            if yNum < 2:
                raise TypeError('Region is underconstrained vertically')
            self.constraints = [
                self._makeXConstraint(xConstraints),
                self._makeYConstraint(yConstraints)]
        else:
            if xNum + yNum > 3:
                raise TypeError('Region is overconstrained')
            if xNum + yNum < 3:
                raise TypeError('Region is underconstrained')
            if xNum == 2:
                self.constraints = self._makeYAspectConstraints(
                    xConstraints, yConstraints, aspect)
            else:
                self.constraints = self._makeXAspectConstraints(
                    xConstraints, yConstraints, aspect)

    def __repr__(self):
        return self._repr

    def _setRepr(self, kwargs):
        self._repr = 'Region(%s)' % (', '.join(
            '%s=%r' % t for t in kwargs.iteritems()))

    def _validateArgNames(self, kwargs):
        for arg in kwargs:
            if arg not in self.VALID_ARGS:
                raise TypeError(
                    'Region() got an unexpected keyword argument ' + repr(arg))

    def getRect(self, app):
        vals = self._evaluateConstraints(app)
        r = pygame.Rect((0, 0), (0, 0))
        self._applyConstraints(r, vals)
        return r

    def _applyConstraints(self, r, vals):
        for c, args in self.constraints:
            c(r, vals, args)

    def _evaluateConstraints(self, app):
        vals = {}
        for k, v, m in self.constraints_2d:
            vals[k] = getattr(v, m)(app)
        for k, v, m in self.constraints_1d:
            result = getattr(v, m)(app)
            vals[k] = (result, result)
        return vals

    def _standardXConstraint(self, r, vals, args):
        p1, p2 = args
        left, width = self._xInterval(vals[p1][0], vals[p2][0])
        r.width = width
        r.left = left

    def _standardYConstraint(self, r, vals, args):
        p1, p2 = args
        top, height = self._yInterval(vals[p1][1], vals[p2][1])
        r.height = height
        r.top = top

    def _makeXConstraint(self, xConstraints):
        self._xInterval, args = self._getIntervalFunction(xConstraints)
        return self._standardXConstraint, args

    def _makeYConstraint(self, yConstraints):
        self._yInterval, args = self._getIntervalFunction(yConstraints)
        return self._standardYConstraint, args

    def _getIntervalFunction(self, constraintsDict):
        if '0' in constraintsDict:
            if 's' in constraintsDict:
                intervalFunc = self._getIntervalFromInterval
                args = constraintsDict['0'], constraintsDict['s']
            elif '2' in constraintsDict:
                intervalFunc = self._getIntervalFromEdges
                args = constraintsDict['0'], constraintsDict['2']
            else:
                intervalFunc = self._getIntervalFromFirstHalf
                args = constraintsDict['0'], constraintsDict['1']
        elif '2' in constraintsDict:
            if 's' in constraintsDict:
                intervalFunc = self._getIntervalFromReverseInterval
                args = constraintsDict['2'], constraintsDict['s']
            else:
                intervalFunc = self._getIntervalFromSecondHalf
                args = constraintsDict['1'], constraintsDict['2']
        else:
            intervalFunc = self._getIntervalFromCentre
            args = constraintsDict['1'], constraintsDict['s']
        return intervalFunc, args

    def _xAspectConstraint(self, r, vals, args):
        aspect, attr, key = args
        r.width = int(r.height * aspect + 0.5)
        setattr(r, attr, vals[key][0])

    def _yAspectConstraint(self, r, vals, args):
        aspect, attr, key = args
        r.height = int(r.width / aspect + 0.5)
        setattr(r, attr, vals[key][1])

    def _makeXAspectConstraints(self, xConstraints, yConstraints, aspect):
        key, val = xConstraints.items().pop()
        if key == '0':
            attr = 'left'
        elif key == '1':
            attr = 'centerx'
        elif key == '2':
            attr = 'right'
        else:
            raise TypeError('Region is overconstrained')
        return [
            self._makeYConstraint(yConstraints),
            (self._xAspectConstraint, (aspect, attr, val))
        ]

    def _makeYAspectConstraints(self, xConstraints, yConstraints, aspect):
        key, val = yConstraints.items().pop()
        if key == '0':
            attr = 'top'
        elif key == '1':
            attr = 'centery'
        elif key == '2':
            attr = 'bottom'
        else:
            raise TypeError('Region is overconstrained')
        return [
            self._makeXConstraint(xConstraints),
            (self._yAspectConstraint, ((aspect + 0.), attr, val))
        ]

    def _getIntervalFromEdges(self, x1, x2):
        return (x1, x2 - x1)

    def _getIntervalFromInterval(self, x1, deltaX):
        return (x1, deltaX)

    def _getIntervalFromFirstHalf(self, x1, midx):
        return (x1, 2 * (midx - x1))

    def _getIntervalFromSecondHalf(self, midx, x2):
        deltaX = 2 * (x2 - midx)
        return (x2 - deltaX, deltaX)

    def _getIntervalFromCentre(self, midx, deltaX):
        return midx - int(deltaX / 2), deltaX

    def _getIntervalFromReverseInterval(self, x2, deltaX):
        return (x2 - deltaX, deltaX)


class PaddedRegion(object):
    def __init__(self, region, padding):
        self.padding = padding
        self.region = region

    def getRect(self, app):
        rect = pygame.Rect(self.region.getRect(app))
        padding = self.padding.getVal(app)
        rect.width += 2 * padding
        rect.height += 2 * padding
        rect.top -= padding
        rect.left -= padding
        return rect


class Abs(object):
    def __init__(self, *args):
        self.values = args

    def __repr__(self):
        return 'Abs(%s)' % ', '.join(repr(v) for v in self.values)

    def getVal(self, app):
        assert len(self.values) == 1
        return self.values[0]
    getXVal = getYVal = getVal
    getXPoint = getYPoint = getVal

    def getPoint(self, app):
        assert len(self.values) == 2
        return self.values
    getSize = getPoint

    def getRect(self, app):
        assert len(self.values) == 4
        return pygame.Rect(*self.values)


class Canvas(object):
    def __init__(self, *args):
        self.values = args

    def __repr__(self):
        return 'Canvas(%s)' % ', '.join(repr(v) for v in self.values)

    def getVal(self, app):
        assert len(self.values) == 1
        return canvasIntervalToScreen(app, self.values[0])
    getXVal = getYVal = getVal

    def getXPoint(self, app):
        assert len(self.values) == 1
        sm = app.screenManager
        return int(self.value[0] * sm.scaleFactor + sm.offsets[0])

    def getYPoint(self, app):
        assert len(self.values) == 1
        sm = app.screenManager
        return int(self.value[1] * sm.scaleFactor + sm.offsets[1])

    def getPoint(self, app):
        assert len(self.values) == 2
        return app.screenManager.placePoint(self.values)

    def getSize(self, app):
        assert len(self.values) == 2
        return app.screenManager.scale(self.values)

    def getRect(self, app):
        assert len(self.values) == 4
        x, y = app.screenManager.size
        xr, yr, wr, hr = self.values
        return pygame.Rect(x * xr, y * yr, x * wr, y * hr)


class CanvasX(object):
    '''Uses a scaled value for the X axis only'''

    def __init__(self, *args):
        self.values = args

    def __repr__(self):
        return 'CanvasX(%s)' % ', '.join(repr(v) for v in self.values)

    def getVal(self, app):
        assert len(self.values) == 1
        return canvasIntervalToScreen(app, self.values[0])
    getXVal = getVal

    def getYVal(self, app):
        assert len(self.values) == 1
        return self.values[0]

    def getPoint(self, app):
        assert len(self.values) == 2
        sm = app.screenManager
        return (int(self.values[0] * sm.scaleFactor + sm.offsets[0]),
                self.values[1])

    def getSize(self, app):
        assert len(self.values) == 2
        sm = app.screenManager
        return (int(self.values[0] * sm.scaleFactor), self.values[1])

    def getRect(self, app):
        assert len(self.values) == 4
        x, y = app.screenManager.size
        xr, yr, wr, hr = self.values
        return pygame.Rect(x * xr, yr, x * wr, hr)


class Screen(object):
    def __init__(self, *args):
        self.values = args

    def __repr__(self):
        return 'Screen(%s)' % ', '.join(repr(v) for v in self.values)

    def getXVal(self, app):
        assert len(self.values) == 1
        return app.screenManager.size[0] * self.values[0]
    getXPoint = getXVal

    def getYVal(self, app):
        assert len(self.values) == 1
        return app.screenManager.size[1] * self.values[1]
    getYPoint = getYVal

    def getPoint(self, app):
        assert len(self.values) == 2
        x, y = app.screenManager.size
        xr, yr = self.values
        return (x * xr, y * yr)
    getSize = getPoint

    def getRect(self, app):
        assert len(self.values) == 4
        x, y = app.screenManager.size
        xr, yr, wr, hr = self.values
        return pygame.Rect(x * xr, y * yr, x * wr, y * hr)


class Relative(object):
    '''
    Provides a point relative to an existing rect.
    @param rectGiver: a function returning a pygame.Rect.
    @param x: relative x value from 0 to 1.
    @param y: relative y value from 0 to 1.
    '''

    def __init__(self, rectGiver, x, y):
        self.rectGiver = rectGiver
        self.x = x
        self.y = y

    def __repr__(self):
        return 'Relative(%r, %r, %r)' % (self.rectGiver, self.x, self.y)

    def getPoint(self, app):
        r = self.rectGiver()
        return (r.left + self.x * r.width, r.top + self.y * r.height)

    def getSize(self, app):
        r = self.rectGiver()
        return (self.x * r.width, self.y * r.height)


class ScaledSize(Canvas):
    '''
    For compatibility with dodgy code like gui.framework.slider.Slider() that
    directly uses .val() of something that it's not sure is a
    ScaledSize/ScaledPoint.
    '''

    def __init__(self, *args):
        Canvas.__init__(self, *args)
        self.val = args

ScaledPoint = ScaledSize

Size = Abs



from collections import defaultdict
import datetime
import json
import logging
import os

from trosnoth import dbqueue
from trosnoth.network.networkDefines import serverVersion
from trosnoth.utils.twist import WeakLoopingCall
from trosnoth.utils.utils import timeNow    # TODO: use game not real time

log = logging.getLogger(__name__)


# TODO: test the accuracy etc. achievements after fixing the stats to fit the
# event-based architecture

class PlayerStatKeeper(object):
    '''Maintains the statistics for a particular player object'''
    def __init__(self, gameRecorder, player):
        self.gameRecorder = gameRecorder
        self.player = player

        # A: Recorded [A]ll game (including post-round)
        # G: Recorded during the main [G]ame only
        # P: Recorded [P]ost-Game only
        self.kills = 0           # G Number of kills they've made
        self.deaths = 0          # G Number of times they've died
        self.zoneTags = 0        # G Number of zones they've tagged
        self.zoneAssists = 0     # G Number of zones they've been in when their
                                 #   team tags it
        self.shotsFired = 0      # A Number of shots they've fired
        self.shotsHit = 0        # A Number of shots that have hit a target
        self.starsEarned = 0     # G Aggregate total of stars earned
        self.starsUsed = 0       # G Aggregate total of stars used
        self.starsWasted = 0     # G Aggregate total of stars died with
        self.roundsWon = 0       # . Number of rounds won
        self.roundsLost = 0      # . Number of rounds lost

        self.playerKills = defaultdict(int)    # A Number of kills against
                                               #   individual people
        self.playerDeaths = defaultdict(int)   # A Number of deaths from
                                               #   individual people
        self.upgradesUsed = defaultdict(int)   # G Number of each upgrade used

        self.timeAlive = 0.0     # G Total time alive
        self.timeDead = 0.0      # G Total time dead

        self.killStreak = 0      # G Number of kills made without dying
        self.currentKillStreak = 0
        self.tagStreak = 0       # G Number of zones tagged without dying
        self.currentTagStreak = 0
        self.aliveStreak = 0.0   # G Greatest time alive in one life
        self.lastTimeRespawned = None
        self.lastTimeDied = None
        self.lastTimeSaved = None

        self.player.onStarsSpent.addListener(self.starsSpent)
        self.player.onShotFired.addListener(self.shotFired)
        self.player.onDied.addListener(self.died)
        self.player.onRespawned.addListener(self.respawned)
        self.player.onKilled.addListener(self.killed)
        self.player.onShotHitSomething.addListener(self.shotHit)
        self.player.onUsedUpgrade.addListener(self.upgradeUsed)

    def stop(self):
        self.player.onStarsSpent.removeListener(self.starsSpent)
        self.player.onShotFired.removeListener(self.shotFired)
        self.player.onDied.removeListener(self.died)
        self.player.onRespawned.removeListener(self.respawned)
        self.player.onKilled.removeListener(self.killed)
        self.player.onShotHitSomething.removeListener(self.shotHit)
        self.player.onUsedUpgrade.removeListener(self.upgradeUsed)

    def shotFired(self, *args, **kwargs):
        self.shotsFired += 1

        if self.shotsFired >= 100:
            accuracy = self.accuracy()
            if accuracy >= 0.10:
                self.sendAchievementProgress('accuracySmall')
            if accuracy >= 0.15:
                self.sendAchievementProgress('accuracyMedium')
            if accuracy >= 0.20:
                self.sendAchievementProgress('accuracyLarge')

        if self.totalPoints() >= 1337:
            self.sendAchievementProgress('statScore')

    def sendAchievementProgress(self, achievementId):
        if self.player.id == -1:
            # Player is no longer in game.
            return
        if not self.gameRecorder:
            return
        self.gameRecorder.game.achievementManager.triggerAchievement(
            self.player, achievementId)

    def _updateStreaks(self, updateAlive):
        '''
        updateAlive will be set to True in three situations:
          1. if the player has just died
          2. if the player was alive when the game ended
          3. if the player was alive when they disconnected
        '''
        self.killStreak = max(self.killStreak, self.currentKillStreak)
        self.tagStreak = max(self.tagStreak, self.currentTagStreak)

        time = timeNow()

        if updateAlive and self.lastTimeRespawned:
            lastLife = time - self.lastTimeRespawned
            self.aliveStreak = max(self.aliveStreak, lastLife)
            self.timeAlive += lastLife
            if lastLife >= 180:
                self.sendAchievementProgress('aliveStreak')

        elif self.lastTimeDied is not None:
            self.timeDead += time - self.lastTimeDied

        self.currentKillStreak = 0
        self.currentTagStreak = 0

    def died(self, killer, deathType):
        self.deaths += 1
        self.playerDeaths[killer] += 1
        self.starsWasted += self.player.starsAtLastDeath

        time = timeNow()

        self._updateStreaks(True)

        self.lastTimeDied = time

        if self.timeAlive >= 1000:
            self.sendAchievementProgress('totalAliveTime')

        if self.timeAlive >= 300 and self.timeAlive >= (self.timeAlive +
                self.timeDead) * 0.75:
            self.sendAchievementProgress('stayingAlive')

    def respawned(self):
        time = timeNow()

        if self.lastTimeDied is not None:
            self.timeDead += (time - self.lastTimeDied)

        self.lastTimeRespawned = time

    def goneFromGame(self):
        self._updateStreaks(not self.player.dead)

    def gameOver(self, winningTeam):
        self._updateStreaks(not self.player.dead)
        if winningTeam is None:
            # Draw. Do nothing
            pass
        elif self.player.isEnemyTeam(winningTeam):
            self.roundsLost += 1
        else:
            self.roundsWon += 1

    def killed(self, victim, deathType):
        self.kills += 1
        self.playerKills[victim] += 1
        self.currentKillStreak += 1
        self.starsEarned += 1

        self.killStreak = max(self.killStreak, self.currentKillStreak)

    def zoneTagged(self):
        self.zoneTags += 1
        self.currentTagStreak += 1
        self.starsEarned += 1

        self.tagStreak = max(self.tagStreak, self.currentTagStreak)

    def tagAssist(self):
        self.zoneAssists += 1

    def shotHit(self, *args, **kwargs):
        self.shotsHit += 1

    def starsSpent(self, stars):
        self.starsUsed += stars

        if self.starsUsed >= 20 and self.starsUsed >= self.starsEarned * 0.5:
            self.sendAchievementProgress('useStarsEfficiently')

    def upgradeUsed(self, upgrade):
        self.upgradesUsed[upgrade.upgradeType] += 1

    def totalPoints(self):
        points = 0
        points += self.kills        * 10
        points += self.deaths       * 1
        points += self.zoneTags     * 20
        points += self.zoneAssists  * 5
        points += self._accuracyPoints()

        return points

    def _accuracyPoints(self):
        if self.shotsFired == 0:
            return 0
        return ((self.shotsHit ** 2.) / self.shotsFired) * 30

    def accuracy(self):
        if self.shotsFired == 0:
            return 0
        return self.shotsHit * 1. / self.shotsFired

    def statDict(self):
        stats = {}
        for val in ('aliveStreak', 'deaths', 'killStreak', 'kills',
                'roundsLost', 'roundsWon', 'shotsFired', 'shotsHit',
                'starsEarned', 'starsUsed', 'starsWasted', 'tagStreak',
                'timeAlive', 'timeDead', 'upgradesUsed', 'zoneAssists',
                'zoneTags'):
            stats[val] = getattr(self, val)
        stats['bot'] = self.player.bot
        stats['team'] = self.player.teamId
        stats['username'] = (self.player.user.username if self.player.user
                else None)

        # We've stored these dicts as player objects (meaning rejoins may be
        # credited in two places)
        # Here, we combine them by aggregating on player nick (which should be
        # the same)
        for attribute in 'playerKills', 'playerDeaths':
            dictionary = getattr(self, attribute)
            newDict = {}
            for player, value in dictionary.iteritems():
                if player is None:
                    continue
                newDict[player.nick] = value
            stats[attribute] = newDict

        return stats

    def rejoined(self, player):
        self.player = player


class StatKeeper(object):

    def __init__(self, gameRecorder, world, filename):
        self.gameRecorder = gameRecorder
        self.world = world
        self.filename = filename
        # In case the server dies prematurely, it's nice
        # To at least have the file there so that
        # future games don't accidentally point to this one.
        if filename:
            with open(self.filename, 'w') as f:
                f.write('{}')

        # A mapping of player ids to statLists
        # (Contains only players currently in the game)
        self.playerStatList = {}
        # A list of all statLists
        # (regardless of in-game status)
        self.allPlayerStatLists = {}
        self.winningTeamId = None

        self.world.onZoneTagged.addListener(self.zoneTagged)
        self.world.onPlayerAdded.addListener(self.playerAdded)
        self.world.onPlayerRemoved.addListener(self.playerRemoved)
        self.world.onGameOver.addListener(self.gameOver)

        for player in self.world.players:
            self.playerAdded(player)

        self.loop = WeakLoopingCall(self, 'save')
        self.loop.start(3)

    def stop(self):
        self.save(force=True)
        self.loop.stop()
        self.world.onZoneTagged.removeListener(self.zoneTagged)
        self.world.onPlayerAdded.removeListener(self.playerAdded)
        self.world.onPlayerRemoved.removeListener(self.playerRemoved)
        self.world.onGameOver.removeListener(self.gameOver)
        for playerStats in self.allPlayerStatLists.itervalues():
            playerStats.stop()

    def save(self, force=False):
        if not force and self.world.gameIsInProgress():
            return
        if not self.filename:
            return
        stats = {}
        stats['players'] = {}
        for playerStat in self.allPlayerStatLists.itervalues():
            stats['players'][playerStat.player.nick] = playerStat.statDict()
        if self.winningTeamId != None:
            stats['winningTeamId'] = self.winningTeamId
        with open(self.filename, 'w') as f:
            json.dump(stats, f, indent=4)

    def zoneTagged(self, zone, player, previousOwner):
        if player is None:
            return
        self.playerStatList[player.id].zoneTagged()
        for assistant in zone.players:
            if assistant.dead:
                continue
            if assistant.team == player.team and assistant != player:
                self.playerStatList[assistant.id].tagAssist()

    def playerAdded(self, player):
        statkeeper = self.allPlayerStatLists.get(player.identifyingName)
        if statkeeper:
            statkeeper.rejoined(player)
        else:
            statkeeper = PlayerStatKeeper(self.gameRecorder, player)
            self.allPlayerStatLists[player.identifyingName] = statkeeper
        self.playerStatList[player.id] = statkeeper

    def playerRemoved(self, player, oldId):
        self.playerStatList[oldId].goneFromGame()
        # Just remove this from the list of current players
        # (retain in list of all stats)
        del self.playerStatList[oldId]

    def gameOver(self, team, timeOver):
        self.winningTeamId = team.id
        # Only credit current players for game over
        for playerStat in self.playerStatList.values():
            playerStat.gameOver(team)
        self.save(force=True)


class ServerGameStats(object):
    def __init__(self, game):
        from trosnoth.djangoapp.models import GameRecord

        self.game = game
        self.world = game.world
        self.gameRecord = GameRecord(
            started=datetime.datetime.now(),
            mapHeight=self.world.map.layout.mapHeight,
            halfMapWidth=self.world.map.layout.halfMapWidth,
            serverVersion=serverVersion,
            blueTeamName=self.world.teams[0].teamName,
            redTeamName=self.world.teams[1].teamName,
            replayName=os.path.basename(game.gameRecorder.replayFilename),
        )
        self.statKeeper = StatKeeper(None, self.world, None)

    def stopAndSave(self):
        '''
        Saves the game stats to the server database.
        '''
        from trosnoth.djangoapp.models import (
            TrosnothUser, GamePlayer, PlayerKills, UpgradesUsedInGameRecord,
        )

        self.statKeeper.stop()
        for playerStat in self.statKeeper.playerStatList.values():
            playerStat._updateStreaks(not playerStat.player.dead)

        self.gameRecord.finished = datetime.datetime.now()
        self.gameRecord.gameSeconds = self.world.timer.getElapsedTime()
        winner = self.world.getWinningTeam()
        self.gameRecord.winningTeam = winner.id if winner else ''
        dbqueue.add(self.gameRecord.save)

        playerRecords = {}
        for playerStat in self.statKeeper.allPlayerStatLists.itervalues():
            player = playerStat.player
            if player.user:
                user = TrosnothUser.fromUser(
                    username=player.user.username)
                bot = False
                botName = ''
            else:
                user = None
                bot = player.bot
                botName = player.nick

            record = GamePlayer(
                game=self.gameRecord,
                user=user, bot=bot, botName=botName,
                team=player.team.id if player.team else '',

                starsEarned=playerStat.starsEarned,
                starsWasted=playerStat.starsWasted,
                starsUsed=playerStat.starsUsed,
                kills=playerStat.kills,
                deaths=playerStat.deaths,
                zoneTags=playerStat.zoneTags,
                zoneAssists=playerStat.zoneAssists,
                shotsFired=playerStat.shotsFired,
                shotsHit=playerStat.shotsHit,
                timeAlive=playerStat.timeAlive,
                timeDead=playerStat.timeDead,
                killStreak=playerStat.killStreak,
                tagStreak=playerStat.tagStreak,
                aliveStreak=playerStat.aliveStreak,
            )
            self.queueSaveWithAttrs(record, ['game'])
            playerRecords[player.identifyingName] = record

            for upgradeType, count in playerStat.upgradesUsed.iteritems():
                upgradeRecord = UpgradesUsedInGameRecord(
                    gamePlayer=record,
                    upgrade=upgradeType,
                    count=count,
                )
                self.queueSaveWithAttrs(upgradeRecord, ['gamePlayer'])

        for playerStat in self.statKeeper.allPlayerStatLists.itervalues():
            killeeRecord = playerRecords[playerStat.player.identifyingName]
            killEntries = {}
            for killer, count in playerStat.playerDeaths.iteritems():
                if killer:
                    killerRecord = playerRecords[killer.identifyingName]
                else:
                    killerRecord = None

                if killer in killEntries:
                    # Same killer after disconnect / reconnect
                    killRecord = killEntries[killer]
                else:
                    killRecord = PlayerKills(
                        killer=killerRecord,
                        killee=killeeRecord,
                    )
                    killEntries[killer] = killRecord

                killRecord.count += count

            for killRecord in killEntries.itervalues():
                self.queueSaveWithAttrs(killRecord, ['killee', 'killer'])

    def queueSaveWithAttrs(self, record, attrs):
        '''
        The Django ORM seems to store the ID of foreign key relationships when
        the attribute is first set, so if you set the relationship before the
        related object is first saved, saving will break. Getting and setting
        again just before the save is good enough to make Django happy.
        '''
        @dbqueue.add
        def saveEndOfGameStats():
            for attr in attrs:
                setattr(record, attr, getattr(record, attr))
            record.save()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import framework, statList, elements, hotkey
from trosnoth.gui.common import Location
from trosnoth.gui.fonts.font import Font
import pygame
from random import randint, shuffle

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        backdrop = pygame.Surface(app.screenManager.scaledSize)
        backdrop.fill((128,0,128))
        self.val = 4
        self.l = ['hello', self.val]
        getColour = lambda item: ((255,0,0), (0,255,255))[str(item).startswith('h')]
        font = Font("KLEPTOCR.TTF", 30)
        self.s = statList.StatList(app,
                              Location((550,400), 'bottomright'),
                              self.l,
                              font,
                              colourFunction = getColour,
                                   align='right')
    
        self.s.setVisible()

        f = lambda: self.s.setVisible(not self.s.getVisibility())

        h = hotkey.Hotkey(app, pygame.K_s, 0)
        h.onTriggered.addListener(f)
        self.elements = [elements.PictureElement(app, backdrop),
                         self.s,
                         h]
    def tick(self, deltaT):
        super(Interface, self).tick(deltaT)
        if randint(0,100) == 1:
            shuffle(self.l)
        if randint(0,30000) == 42:
            self.l.append("What's going on here?!")
        if randint(0,1000) == 100:
            i = self.l.index(self.val)
            self.val = randint(0, 1000)
            self.l[i] = self.val

size = (600,450)

if __name__ == '__main__':
    a = Application(size, 0, "Testing", Interface)
    a.run()



from ctypes import (
    Structure, c_uint32, c_float, c_wchar, c_char, sizeof, c_char_p, CDLL,
)
import mmap
import os
import sys

import simplejson


WIN32 = sys.platform == 'win32'

if WIN32:
    from ctypes import windll
    FILE_MAP_ALL_ACCESS = 0xF001F
    FALSE = 0
else:
    try:
        librt = CDLL('librt.so')    # Linux
    except:
        try:
            librt = CDLL('/usr/lib/libSystem.dylib')    # OSX
        except:
            librt = None
    S_IRUSR = 0o400
    S_IWUSR = 0o200


class MumbleLinkedMem(Structure):
    _fields_ = [
        ('uiVersion', c_uint32),
        ('uiTick', c_uint32),
        ('fAvatarPosition', c_float * 3),
        ('fAvatarFront', c_float * 3),
        ('fAvatarTop', c_float * 3),
        ('name', c_wchar * 256),
        ('fCameraPosition', c_float * 3),
        ('fCameraFront', c_float * 3),
        ('fCameraTop', c_float * 3),
        ('identity', c_wchar * 256),
        ('context_len', c_uint32),
        ('context', c_char * 256),
        ('description', c_wchar * 2048),
    ]


class MumbleUpdater(object):
    mapUnitsPerMetre = 1000

    def __init__(self):
        self.lm = None
        self.mapScale = 1. / self.mapUnitsPerMetre
        self.lastPlayerDetails = None

        if WIN32:
            # TODO: test this on windows
            hMapObject = windll.kernel32.OpenFileMappingW(
                FILE_MAP_ALL_ACCESS, FALSE, u'MumbleLink')
            if hMapObject == 0:
                return

            address = windll.kernel32.MapViewOfFile(
                hMapObject, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(MumbleLinkedMem))
            if address == 0:
                windll.kernel32.CloseHandle(hMapObject)
                return
            self.lm = MumbleLinkedMem.from_address(address)

        elif librt:
            memname = c_char_p('/MumbleLink.%d' % (os.getuid(),))
            shmfd = librt.shm_open(memname, os.O_RDWR, S_IRUSR | S_IWUSR)
            if shmfd < 0:
                # Shared memory does not exist, because Mumble is not running
                return

            self.lm = MumbleLinkedMem.from_buffer(
                mmap.mmap(shmfd, sizeof(MumbleLinkedMem)))

    def scalePoint(self, (x, y)):
        return (x * self.mapScale, -y * self.mapScale, 0)

    def update(self, player, cameraPos, serverId):
        if self.lm is None:
            return

        if self.lm.uiVersion != 2:
            self.lm.name = 'Trosnoth'
            self.lm.description = 'Trosnoth network platformer'
            self.lm.uiVersion = 2
            self.lastPlayerDetails = None

        self.lm.uiTick += 1

        self.lm.fAvatarFront = self.lm.fCameraFront = (0, 0, 1)
        self.lm.fAvatarTop = self.lm.fCameraTop = (0, 1, 0)
        if player:
            self.lm.fAvatarPosition = self.scalePoint(player.pos)
            playerDetails = (player.teamId, player.id)
        else:
            self.lm.fAvatarPosition = (0, 0, 0)
            playerDetails = []

        self.lm.fCameraPosition = self.scalePoint(cameraPos)

        self.lm.context = serverId
        self.lm.context_len = len(serverId)

        if playerDetails != self.lastPlayerDetails:
            self.lastPlayerDetails = playerDetails
            self.lm.identity = simplejson.dumps(playerDetails)


mumbleUpdater = MumbleUpdater()



import logging

from trosnoth.messages.base import AgentRequest, ServerCommand

log = logging.getLogger()


class PlayerHasElephantMsg(ServerCommand):
    idString = 'jken'
    fields = 'playerId'
    packspec = 'c'


class PlayerHasTrosballMsg(ServerCommand):
    idString = 'ball'
    fields = 'playerId'
    packspec = 'c'


class TrosballPositionMsg(ServerCommand):
    idString = 'bing'
    fields = 'xpos', 'ypos', 'xvel', 'yvel'
    packspec = 'ffff'


class ThrowTrosballMsg(AgentRequest):
    idString = 'pass'
    fields = 'tickId'
    packspec = 'H'

    def clientValidate(self, localState, world, sendResponse):
        if not localState.player:
            return False
        if localState.player.id != world.trosballPlayer.id:
            return False
        return True

    def serverApply(self, game, agent):
        if agent.player and agent.player == game.world.trosballPlayer:
            game.world.throwTrosball()


class TrosballTagMsg(ServerCommand):
    idString = 'goal'
    fields = 'teamId', 'playerId'
    packspec = 'cc'


class AchievementUnlockedMsg(ServerCommand):
    idString = 'Achm'
    fields = 'playerId', 'achievementId'
    packspec = 'c*'



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import random
import logging

from trosnoth.messages import RemovePlayerMsg
from trosnoth.const import DEFAULT_TEAM_NAME_1, DEFAULT_TEAM_NAME_2

log = logging.getLogger(__name__)


class VoteArbiter(object):
    def __init__(self, universe, gameStartMethod):
        self.universe = universe
        self.gameStartMethod = gameStartMethod

    def getPlayers(self):
        return [p for p in self.universe.players if not p.bot]

    def getTeams(self):
        return self.universe.teams

    def startNewGameIfReady(self):
        if self.readyForTournament():
            self.startTournamentGame()
        elif self.readyForScratchMatch():
            self.startNewGame()

    def getReadyTournamentPlayersByTeam(self):
        tournamentTeams = {}
        for p in self.getPlayers():
            if p.prefersTournamentTeam and p.readyForTournament:
                tournamentTeams.setdefault(p.preferredTeam, []).append(p)
        return tournamentTeams

    def getReadyTournamentTeams(self):
        if not self.universe.authManager:
            return []

        tournamentTeams = self.getReadyTournamentPlayersByTeam()

        minReady = self.universe.authManager.tournamentSettings.minTeamSize
        return [t for t, players in tournamentTeams.items() if
                len(players) >= minReady]

    def readyForTournament(self):
        readyTeams = self.getReadyTournamentTeams()

        return len(readyTeams) >= 2

    def readyForScratchMatch(self):
        players = self.getPlayers()
        totalPlayers = len(players)

        readyPlayerCount = len([p for p in players if p.readyToStart])

        if totalPlayers < 2:
            return False
        return readyPlayerCount >= 0.7 * totalPlayers

    def startTournamentGame(self):
        players = self.getReadyTournamentPlayersByTeam()
        teams = self.getReadyTournamentTeams()

        while len(teams) > 2:
            teams.pop(random.randrange(len(teams)))

        settings = self.universe.authManager.tournamentSettings
        size = settings.mapSize
        duration = settings.gameDuration
        maxTeamSize = settings.maxTeamSize

        teamName1, teamName2 = teams
        players1 = players[teamName1][:maxTeamSize]
        players2 = players[teamName2][:maxTeamSize]

        log.info('Starting tournament match with %r vs %r', teamName1,
                teamName2)
        log.info('Duration %r, size %r', duration, size)
        self.assignTeamNames(teamName1, teamName2)
        self.assignPlayersToTeams(players1, players2)
        self.bootPlayers(p for p in self.getPlayers() if p not in players1 and p
                not in players2)

        self.gameStartMethod(duration, size)

    def assignTeamNames(self, teamName1, teamName2):
        # Set team names.
        self.getTeams()[0].teamName = teamName1
        self.getTeams()[1].teamName = teamName2

    def assignPlayersToTeams(self, players1, players2):
        for i, players in [(0, players1), (1, players2)]:
            team = self.getTeams()[i]
            for player in players:
                player.team = team

    def bootPlayers(self, players):
        for player in players:
            self.universe.sendServerCommand(RemovePlayerMsg(player.id))

    def startNewGame(self):
        result = self._getNewTeams()
        if result is None:
            return
        teamName1, players1, teamName2, players2 = result

        size = self._getNewSize(min(len(players1), len(players2)))
        duration = self._getNewDuration(size)

        self.assignTeamNames(teamName1, teamName2)
        self.assignPlayersToTeams(players1, players2)
        self.gameStartMethod(duration, size)

    def _getNewDuration(self, mapSize):
        results = {}
        for player in self.getPlayers():
            duration = player.preferredDuration
            results[duration] = results.get(duration, 0) + 1
        items = results.items()
        items.sort(key=lambda (duration, count): count)
        items.reverse()
        duration, count = items[0]
        if results.get(0, 0) != count:
            return duration

        # Decide default based on map size.
        if mapSize == (3, 2):
            return 45 * 60
        elif mapSize == (1, 1):
            return 10 * 60
        elif mapSize == (5, 1):
            return 20 * 60
        else:
            return min(7200, 2 * 60 * (mapSize[0]*2+1) * (mapSize[1]*1.5))

    def _getNewSize(self, teamSize):
        '''
        Returns a new map size based on what players vote for, with the
        defaults (if most people select Auto) being determined by the size of
        the smaller team.
        '''
        results = {}
        for player in self.getPlayers():
            size = player.preferredSize
            results[size] = results.get(size, 0) + 1
        items = results.items()
        items.sort(key=lambda (size, count): count)
        items.reverse()
        size, count = items[0]
        if results.get((0, 0), 0) != count:
            return size

        # Decide size based on player count.
        if teamSize <= 3:
            return (1, 1)
        elif teamSize <= 4:
            return (5, 1)
        else:
            return (3, 2)

    def _getNewTeams(self):
        '''
        Returns (teamName1, players1, teamName2, players2) based on what teams
        people have selected as their preferred teams. Bots will not be put on
        any team.
        '''
        teamName1, players1, teamName2, players2, others = (
                self._getRelevantTeamPreferences())

        totalPlayers = len(self.getPlayers())
        fairLimit = (totalPlayers + 1) // 2
        if len(players1) == totalPlayers:
            # Don't start if everyone's on one team.
            return None

        if len(players1) > fairLimit:
            # Require every player on the disadvantaged team to be ready.
            for player in players2 + others:
                if not player.readyToStart:
                    return None

        random.shuffle(others)
        for player in others:
            count1 = len(players1)
            count2 = len(players2)
            if count1 > count2:
                players2.append(player)
            elif count2 > count1:
                players1.append(player)
            else:
                random.choice([players1, players2]).append(player)

        return teamName1, players1, teamName2, players2

    def _getRelevantTeamPreferences(self):
        '''
        Returns (teamName1, players1, teamName2, players2, otherPlayers) based
        on what teams people have selected as their preferred teams. Players who
        have not selected one of the two most popular teams will be in the
        otherPlayers collection.
        '''
        desiredTeams = self._getDesiredTeams()
        others = []
        if desiredTeams[0][0] == '':
            teamName, players = desiredTeams.pop(0)
            others.extend(players)

        if desiredTeams:
            teamName1, players1 = desiredTeams.pop(0)
        else:
            teamName1, players1 = '', []

        if desiredTeams:
            teamName2, players2 = desiredTeams.pop(0)
        else:
            teamName2, players2 = '', []

        for teamName, players in desiredTeams:
            others.extend(players)

        if teamName1 == '':
            teamName1 = (DEFAULT_TEAM_NAME_1 if teamName2 != DEFAULT_TEAM_NAME_1
                    else DEFAULT_TEAM_NAME_2)
        if teamName2 == '':
            teamName2 = (DEFAULT_TEAM_NAME_1 if teamName1 != DEFAULT_TEAM_NAME_1
                    else DEFAULT_TEAM_NAME_2)

        return teamName1, players1, teamName2, players2, others

    def _getDesiredTeams(self):
        '''
        Returns a sorted sequence of doubles of the form (teamName, players)
        where teamName is a unicode/string and players is a list of players. The
        sequence will be sorted from most popular to least popular.
        '''
        results = {}
        for player in self.getPlayers():
            teamName = player.preferredTeam
            results.setdefault(teamName, []).append(player)
        items = results.items()
        items.sort(key=lambda (teamName, players): (len(players), teamName))
        items.reverse()
        return items



import logging

log = logging.getLogger('framework')


class Element(object):
    active = True
    hasFocus = False

    def __init__(self, app, *args, **kwargs):
        super(Element, self).__init__(*args, **kwargs)
        self.app = app

    def processEvent(self, event):
        '''Processes the specified event and returns the event if it should
        be passed on, or None if it has been caught.'''
        return event

    def tick(self, deltaT):
        '''Gives the element a chance to update itself. deltaT is the time
        in seconds since the last tick cycle.'''

    def draw(self, screen):
        '''Gives the element a chance to draw itself onto the screen.'''

    def gotFocus(self):
        pass

    def lostFocus(self):
        pass


class CompoundElement(Element):
    # List of elements in order that they are drawn on screen. That is,
    # the first element in the list is drawn at the bottom.

    def __init__(self, *args, **kwargs):
        super(CompoundElement, self).__init__(*args, **kwargs)
        self.elements = []
        self.active = True

    def processEvent(self, event):
        if not self.active:
            return event

        elements = list(self.elements)
        elements.reverse()
        for element in elements:
            event = element.processEvent(event)

            if event is None or not self.app._running:
                return None
        return event

    def tick(self, deltaT):
        if not self.active:
            return
        try:
            for element in self.elements:
                try:
                    element.tick(deltaT)
                except Exception, e:
                    log.error(str(e), exc_info=True)
        except RuntimeError, e:
            if e.__str__() == 'maximum recursion depth exceeded':
                log.error('ERROR: Infinite loop of elements:')
                if hasattr(element, 'elements'):
                    log.error('%s contains %s' % (element, element.elements))
            raise

    def draw(self, screen):
        if not self.active:
            return
        for element in self.elements:
            try:
                element.draw(screen)
            except Exception, e:
                log.error('%s: %s', self, e, exc_info=True)

    def clearFocus(self):
        self.setFocus(None)

    def setFocus(self, element):
        if element is not None and element not in self.elements:
            return

        if element is not None and element.hasFocus:
            return

        for el in self.elements:
            if el.hasFocus:
                el.hasFocus = False
                el.lostFocus()
        if element is not None:
            element.hasFocus = True
            element.gotFocus()


class TabFriendlyCompoundElement(CompoundElement):
    def __init__(self, app, *args, **kwargs):
        super(TabFriendlyCompoundElement, self).__init__(app, *args, **kwargs)
        self.tabOrder = []

    def tabNext(self, sender):
        assert sender in self.tabOrder
        i = self.tabOrder.index(sender)
        i += 1
        i %= len(self.tabOrder)
        self.setFocus(self.tabOrder[i])


class SwitchingElement(Element):
    def __init__(self, app, elements=None, choice=None, *args, **kwargs):
        super(SwitchingElement, self).__init__(app, *args, **kwargs)
        if elements is None:
            self.elements = {}
        else:
            self.elements = elements

        self.setChoice(choice)

    def setChoice(self, choice):
        if isinstance(choice, Element):
            self.element = choice
        else:
            self.element = self.elements.get(choice, None)

    def processEvent(self, event):
        if not self.element:
            return event
        return self.element.processEvent(event)

    def tick(self, deltaT):
        if self.element:
            self.element.tick(deltaT)

    def draw(self, screen):
        if self.element:
            self.element.draw(screen)



'''chatBox.py - defines the ChatBox class which deals with drawing the
chat box (the box in which you input chat) to the screen.'''

import logging

import pygame

import trosnoth.gui.framework.framework as framework
from trosnoth.gui.framework.elements import TextElement
from trosnoth.gui.framework.prompt import InputBox
from trosnoth.gui.common import Area, FullScreenAttachedPoint, Location
from trosnoth.trosnothgui.ingame.messagebank import MessageBank
from trosnoth.utils.utils import wrapline

log = logging.getLogger()


class ChatBox(framework.CompoundElement):
    def __init__(self, app, world, interface):
        super(ChatBox, self).__init__(app)

        self.world = world
        self.app = app
        self.interface = interface

        self.font = self.app.screenManager.fonts.newChatFont

        self.frameColour = self.app.theme.colours.chatFrameColour
        self.insideColour = self.app.theme.colours.chatInsideColour
        self.textColour = self.app.theme.colours.chatNormalColour

        self.sayToTeam = TextElement(self.app,
            text="Say to team:",
            font=self.font,
            pos=Location(FullScreenAttachedPoint((20, 501), 'topleft'),
                'topleft'),
            colour=self.textColour,
            shadow=True,
        )

        self.inputPosition = Area(FullScreenAttachedPoint((145, 500),
            'topleft'), (370,20), 'topleft')
        self.input = InputBox(self.app, self.inputPosition, font=self.font)
        self.input.onEnter.addListener(lambda sender:
                self.hitEnter(sender.value))
        self.input.onEsc.addListener(lambda sender: self.close())
        self.input.onClick.addListener(self.setFocus)

        self.messages = MessageBank(self.app, 10, 100,
                Location(FullScreenAttachedPoint((20,470),'topleft'),
                'topleft'), 'left', 'bottom', self.font)

        self._chatOpen = False
        self.teamChat = True
        self.player = None

        self.messageBuffer = []

        self.MESSAGE_GAP = self.font.getHeight(self.app)

        self.elements = [self.messages]

    def setPlayer(self, player):
        self.player = player
        if self.player.team is None:
            self.refreshMode()

    def canTeamChat(self):
        return self.player and self.player.team is not None

    def switchModes(self):
        self.teamChat = not self.teamChat
        self.refreshMode()

    def refreshMode(self):
        self.teamChat = self.teamChat and self.canTeamChat()
        if self.teamChat:
            self.sayToTeam.setText("Say to team:")
        else:
            self.sayToTeam.setText("Say to all:")

    def hitEnter(self, senderValue):
        if senderValue.strip() != '':
            self.sendChat(senderValue)
        self.input.clear()

    def sendChat(self, senderValue):
        # Interpret lines with initial hash.
        if senderValue.startswith('#'):
            i = 1
            while senderValue[i:i+1].isdigit():
                i += 1
            try:
                playerId = chr(int(senderValue[1:i]))
            except ValueError:
                pass
            else:
                self.interface.sendPrivateChat(self.player, playerId,
                        senderValue[i:].lstrip())
                return

        if self.teamChat:
            self.interface.sendTeamChat(self.player, senderValue)
        else:
            self.interface.sendPublicChat(self.player, senderValue)

    def refresh(self):
        if not self.isOpen():
            return
        self.elements = [self.sayToTeam, self.input]

        initialY = 470
        count = 0
        for text, nick, colour, firstLine in reversed(self.messageBuffer):
            currentY = initialY - count * self.MESSAGE_GAP
            if currentY < 200 or count >= 10:
                break

            if firstLine and nick is not None:
                person = TextElement(self.app,
                    text=nick, font=self.font,
                    pos=Location(FullScreenAttachedPoint((20, currentY),
                        'topleft'), 'topleft'),
                    colour=colour,
                    shadow=True,
                )

                xOffset = person._getRect().width
                self.elements.append(person)

                text = text[len(nick):]

            else:
                xOffset = 0

            if nick is None:
                colour = self.app.theme.colours.serverChat
            else:
                colour = self.textColour

            message = TextElement(self.app,
                text=text,
                font=self.font,
                pos=Location(FullScreenAttachedPoint((20 + xOffset,
                    currentY), 'topleft'), 'topleft'),
                colour=colour,
                shadow=True,
            )
            self.elements.append(message)
            count += 1

    def open(self):
        self._chatOpen = True
        self.refresh()
        self.setFocus(self.input)
        self.input.clear()

    def close(self):
        self._chatOpen = False
        self.elements = [self.messages]
        pygame.key.set_repeat()

    def isOpen(self):
        return self._chatOpen

    def newMessage(self, message, nick, colour):

        message = nick + message

        wrappedMessage = wrapline(message, self.font._getFont(self.app), 480)

        # Update the "box open" message elements
        firstLine = True
        for line in wrappedMessage:
            self.messageBuffer.append((line, nick, colour, firstLine))
            firstLine = False

        while len(self.messageBuffer) > 100:
            self.messageBuffer.remove(0)

        # Update the "box closed" message elements
        firstLine = wrappedMessage.pop(0)[len(nick):]
        parts = [(nick, colour), (firstLine, self.textColour)]

        self.messages.newColourMessage(parts)

        for line in wrappedMessage:
            self.messages.newMessage(line, self.textColour)

        self.refresh()

    def newServerMessage(self, message):
        '''Server messages don't follow the format of normal messages and
        require special attention to display.'''

        colour = self.app.theme.colours.serverChat
        wrappedMessage = wrapline("SERVER" + message,
                self.font._getFont(self.app), 480)

        # Update the "box open" message elements
        firstLine = True
        for line in wrappedMessage:
            self.messageBuffer.append((line, None, colour, firstLine))
            firstLine = False

        while len(self.messageBuffer) > 100:
            self.messageBuffer.remove(0)

        # Update the "box closed" message elements
        for line in wrappedMessage:
            self.messages.newMessage(line, colour)

        self.refresh()

    def _getRect(self):
        return self.area.getRect(self.app)

    def draw(self, surface):

        if self._chatOpen:

            pointX = 5
            pointY = 300

            # Draw the frame first
            frame = pygame.Surface((520,230))
            frame.fill(self.frameColour)

            mainBox = pygame.Surface((500, 180))
            mainBox.fill(self.insideColour)

            sayBox = pygame.Surface((120, 20))
            sayBox.fill(self.insideColour)

            if self.app.displaySettings.alphaOverlays:
                mainBox.set_alpha(128)
                sayBox.set_alpha(128)
                frame.set_alpha(128)

            surface.blit(frame, (pointX, pointY))
            surface.blit(mainBox, (pointX + 10, pointY + 10))
            surface.blit(sayBox, (pointX + 10, pointY + 200))

        super(ChatBox, self).draw(surface)

    def processEvent(self, event):
        if not self.isOpen():
            return event

        if event.type == pygame.KEYDOWN and event.key in (pygame.K_LCTRL,
                pygame.K_RCTRL):
            self.switchModes()
        else:
            return self.input.processEvent(event)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

from trosnoth.gui.framework.menu import MenuDisplay
from trosnoth.gui.menu.menu import MenuManager, Menu, MenuItem
from trosnoth.gui.common import Abs, Location, Region, Canvas, ScaledSize
from trosnoth.gui.framework.dialogbox import DialogBox, DialogResult
from trosnoth.gui.framework.elements import (TextButton, SolidRect,
        TextElement)
from trosnoth.gui.framework.framework import CompoundElement
from trosnoth.gui.framework.prompt import InputBox, intValidator
from trosnoth.messages import (PlayerIsReadyMsg, SetPreferredTeamMsg,
        SetPreferredDurationMsg, SetPreferredSizeMsg)
from trosnoth.utils.event import Event

log = logging.getLogger(__name__)

MENU_WIDTH = 120
class GameVoteMenu(CompoundElement):
    def __init__(self, app, world, onChange=None):
        CompoundElement.__init__(self, app)
        self.world = world
        self.playerId = '\x00'

        self.readyMenu = ReadyMenu(self, 0)
        self.teamMenu = TeamMenu(self, MENU_WIDTH)
        self.sizeMenu = SizeMenu(self, 2*MENU_WIDTH)
        self.durationMenu = DurationMenu(self, 3*MENU_WIDTH)

        self.onChange = Event()
        if onChange is not None:
            self.onChange.addListener(onChange)

        self.subMenu = GameVoteSubMenu(app)
        font = app.screenManager.fonts.ingameMenuFont
        self.elements = [
            SolidRect(app, (0, 64, 192), None,
                Region(topleft=Abs(0,0), width=Abs(4*MENU_WIDTH),
                height=Abs(font.getHeight(app)+2))),
            self.subMenu,
            self.readyMenu.button,
            self.teamMenu.button,
            self.sizeMenu.button,
            self.durationMenu.button,
        ]

    def update(self, player):
        self.playerId = player.id
        self.readyMenu.updateTitle(player)
        self.teamMenu.updateTitle(player)
        self.sizeMenu.updateTitle(player)
        self.durationMenu.updateTitle(player)
        if self.subMenu.manager.menu is not None:
            self.subMenu.manager.menu.update()

    def showSubMenu(self, menu):
        menu.update()
        if self.subMenu.manager.menu is menu or self.subMenu.hidden:
            self.subMenu.hide()
        self.subMenu.manager.setDefaultMenu(menu)
        self.subMenu.location = menu.button.pos

    def hideSubMenu(self):
        if not self.subMenu.hidden:
            self.subMenu.hide()

class GameVoteSubMenu(MenuDisplay):
    ACCELERATION = 1000     # pix/s/s
    def __init__(self, app, keymapping=None):
        font = app.screenManager.fonts.ingameMenuFont
        titleColour = (255, 255, 255)
        stdColour = (255, 255, 0)
        hvrColour = (0, 255, 255)
        backColour = (0, 64, 192)
        disabledColour = (128, 128, 128)
        autosize = True
        hidable = True
        size = Abs(MENU_WIDTH, 10)

        manager = MenuManager()

        location = Location(Abs(0, 0))
        MenuDisplay.__init__(self, app, location, size, font, manager,
                titleColour, stdColour, hvrColour, disabledColour, backColour,
                autosize, hidable, keymapping)

class CountedMenuItem(MenuItem):
    def __init__(self, caption, action=None, listener=None):
        self.caption = caption
        self._count = 0
        MenuItem.__init__(self, caption, action, listener)

    def _getCount(self):
        return self._count
    def _setCount(self, count):
        self._count = count
        self.name = '%s (%d)' % (self.caption, self._count)
    count = property(_getCount, _setCount)

class SelectionMenu(Menu):
    def __init__(self, gvm, pos, items):
        self.gvm = gvm
        self.selected = 0

        Menu.__init__(self, '', items, self.selectItem)

        app = gvm.app
        stdColour = (255, 255, 0)
        hvrColour = (0, 255, 255)
        backColour = (0, 64, 192)
        self.button = TextButton(app, Location(Abs(pos, 0)), '',
            app.screenManager.fonts.ingameMenuFont, stdColour, hvrColour,
            backColour=backColour, onClick=self._toggleSubMenu)

    def _toggleSubMenu(self, element):
        self.gvm.showSubMenu(self)

    def update(self):
        pass

class ReadyMenu(SelectionMenu):
    def __init__(self, gvm, pos):
        self.scratchItem = CountedMenuItem('Ready', (True, False))
        self.tournamentItem = CountedMenuItem('Tournament', (False, True))
        self.tournamentMode = False
        SelectionMenu.__init__(self, gvm, pos, [
            CountedMenuItem('Not ready', (False, False)),
            self.scratchItem,
        ])

    def selectItem(self, (ready, tournament)):
        self.gvm.onChange.execute(PlayerIsReadyMsg(self.gvm.playerId, ready,
                tournament))
        self.gvm.hideSubMenu()

    def updateTitle(self, player):
        self.tournamentItem.enabled = player.prefersTournamentTeam

        if player.readyToStart:
            if not self.tournamentMode:
                title = 'Ready'
            else:
                title = 'Scratch'
        elif player.readyForTournament:
            title = 'Tournament'
        else:
            title = 'Not ready'
        self.button.setText(title)

    def update(self):
        values = {}
        for player in self.gvm.world.players:
            if not player.bot:
                key = player.readyToStart, player.readyForTournament
                values[key] = values.get(key, 0) + 1

        tournamentMode = (False, True) in values or self.tournamentItem.enabled
        if tournamentMode and not self.tournamentMode:
            self.appendItem(self.tournamentItem)
            self.scratchItem.caption = 'Scratch'
        elif not tournamentMode and self.tournamentMode:
            self.popItem(2)
            self.scratchItem.caption = 'Ready'
        self.tournamentMode = tournamentMode

        for item in self.items:
            item.count = values.get(item.action, 0)

class TeamMenu(SelectionMenu):
    def __init__(self, gvm, pos):
        items = [
            CountedMenuItem('Any team', ''),
            MenuItem('---', None),
            MenuItem('Other...', None, listener=self._otherClicked),
        ]
        SelectionMenu.__init__(self, gvm, pos, items)

    def updateTitle(self, player):
        if player.preferredTeam == '':
            title = 'Any team'
        else:
            title = player.preferredTeam
        self.button.setText(title)

    def _otherClicked(self, item):
        prompt = TeamNameBox(self.gvm.app)
        @prompt.onClose.addListener
        def _customEntered():
            if prompt.result == DialogResult.OK:
                team = prompt.value
                self.gvm.onChange.execute(SetPreferredTeamMsg(team.encode()))

        prompt.show()
        self.gvm.hideSubMenu()

    def selectItem(self, team):
        if team is None:
            return
        self.gvm.onChange.execute(SetPreferredTeamMsg(team.encode()))
        self.gvm.hideSubMenu()

    def update(self):
        values = {}
        for player in self.gvm.world.players:
            if not player.bot:
                team = player.preferredTeam
                values[team] = values.get(team, 0) + 1
        killable_items = []
        for item in self.items:
            if item.action is None:
                continue
            if item.action in values:
                item.count = values.pop(item.action)
            else:
                item.count = 0
                if item.action != '':
                    killable_items.append(item)
        for item in killable_items:
            self.removeItem(item)
        for team, count in values.iteritems():
            item = CountedMenuItem(team, team)
            item.count = count
            self.insertItem(-2, item)

DEFAULT_MAP_SIZES = (
    ((0, 0), 'Auto size map'),
    ((3, 2), 'Stardard map'),
    ((1, 1), 'Small map'),
    ((5, 1), 'Wide map'),
)

MAP_SIZE_NAMES = dict(DEFAULT_MAP_SIZES)

class SizeMenu(SelectionMenu):
    def __init__(self, gvm, pos):
        items = []
        for size, name in DEFAULT_MAP_SIZES:
            items.append(CountedMenuItem(name, size))
        items.extend([
            MenuItem('---', None),
            MenuItem('Other...', None, listener=self._otherClicked),
        ])
        SelectionMenu.__init__(self, gvm, pos, items)

    def updateTitle(self, player):
        title = MAP_SIZE_NAMES.get(player.preferredSize)
        if title is None:
            title = '%d x %d map' % player.preferredSize
        self.button.setText(title)

    def _otherClicked(self, item):
        prompt = MapSizeBox(self.gvm.app)
        @prompt.onClose.addListener
        def _customEntered():
            if prompt.result == DialogResult.OK:
                width, height = prompt.value
                width = max(0, min(10, width))
                height = max(0, min(10, height))
                self.gvm.onChange.execute(SetPreferredSizeMsg(self.gvm.playerId,
                        width, height))

        prompt.show()
        self.gvm.hideSubMenu()

    def selectItem(self, size):
        if size is None:
            return
        self.gvm.onChange.execute(SetPreferredSizeMsg(self.gvm.playerId,
                size[0], size[1]))
        self.gvm.hideSubMenu()

    def update(self):
        values = {}
        for player in self.gvm.world.players:
            if not player.bot:
                size = player.preferredSize
                values[size] = values.get(size, 0) + 1
        killable_items = []
        for item in self.items:
            if item.action is None:
                continue
            if item.action in values:
                item.count = values.pop(item.action)
            else:
                item.count = 0
                if item.action not in MAP_SIZE_NAMES:
                    killable_items.append(item)
        for item in killable_items:
            self.removeItem(item)
        for size, count in values.iteritems():
            item = CountedMenuItem('%d x %d map' % size, size)
            item.count = count
            self.insertItem(-2, item)

class DurationMenu(SelectionMenu):
    def __init__(self, gvm, pos):
        SelectionMenu.__init__(self, gvm, pos, [
            CountedMenuItem('Auto duration', 0),
            MenuItem('---', None),
            MenuItem('Other...', None, listener=self._otherClicked),
        ])

    def updateTitle(self, player):
        duration = player.preferredDuration
        if duration == 0:
            title = 'Auto duration'
        else:
            title = '%d min' % (int(duration/60+0.5),)
        self.button.setText(title)

    def _otherClicked(self, item):
        prompt = GameDurationBox(self.gvm.app)
        @prompt.onClose.addListener
        def _customEntered():
            if prompt.result == DialogResult.OK:
                duration = int(prompt.value) * 60
                self.gvm.onChange.execute(SetPreferredDurationMsg(
                        self.gvm.playerId, duration))

        prompt.show()
        self.gvm.hideSubMenu()


    def selectItem(self, duration):
        if duration is None:
            return
        self.gvm.onChange.execute(SetPreferredDurationMsg(self.gvm.playerId,
                duration))
        self.gvm.hideSubMenu()

    def update(self):
        values = {}
        for player in self.gvm.world.players:
            if not player.bot:
                duration = player.preferredDuration
                values[duration] = values.get(duration, 0) + 1
        killable_items = []
        for item in self.items:
            if item.action is None:
                continue
            if item.action in values:
                item.count = values.pop(item.action)
            else:
                item.count = 0
                if item.action != 0:
                    killable_items.append(item)
        for item in killable_items:
            self.removeItem(item)
        for duration, count in values.iteritems():
            item = CountedMenuItem('%d min' % (int(duration/60+0.5),),
                    duration)
            item.count = count
            self.insertItem(-2, item)

class MapSizeBox(DialogBox):
    def __init__(self, app, width=2, height=1):
        super(MapSizeBox, self).__init__(app, ScaledSize(400,230),
                "Custom Size")
        labelFont = app.screenManager.fonts.bigMenuFont
        labelColour = app.theme.colours.dialogBoxTextColour
        btnFont = app.screenManager.fonts.bigMenuFont
        btnColour = app.theme.colours.dialogButtonColour
        highlightColour = app.theme.colours.black
        inputFont = app.screenManager.fonts.defaultTextBoxFont
        inputColour = app.theme.colours.grey

        self.widthBox = InputBox(app, Region(midleft=self.Relative(0.65, 0.25),
                size=self.Relative(0.15, 0.2)), str(width),
                font=inputFont, colour=inputColour, onClick=self.setFocus,
                onEnter=self.okClicked,
                validator=intValidator, maxLength=2)

        self.heightBox = InputBox(app, Region(midleft=self.Relative(0.65, 0.55),
                size=self.Relative(0.15, 0.2)),
                str(height), font=inputFont, colour=inputColour,
                onClick=self.setFocus, onEnter=self.okClicked,
                validator=intValidator, maxLength=2)

        # Add elements to screen
        self.elements = [
            TextElement(app, 'Half Width:', labelFont,
                Location(self.Relative(0.6, 0.25), 'midright'),
                labelColour),
            self.widthBox,

            TextElement(app, 'Height:', labelFont,
                Location(self.Relative(0.6, 0.55), 'midright'),
                labelColour),
            self.heightBox,

            TextButton(app,
                Location(self.Relative(0.3, 0.9), 'center'),
                'Ok', btnFont, btnColour, highlightColour,
                onClick=self.okClicked),
            TextButton(app,
                Location(self.Relative(0.7, 0.9), 'center'),
                'Cancel', btnFont, btnColour, highlightColour,
                onClick=self.cancelClicked),
            ]

        self.tabOrder = [self.widthBox, self.heightBox]
        self.setFocus(self.widthBox)
        #self.setColours(colours.dialogBoxEdgeColour,
        #        colours.dialogBoxTextColour, colours.dialogBoxTextColour)

    def okClicked(self, element):
        self.result = DialogResult.OK
        self.value = (int(self.widthBox.value), int(self.heightBox.value))
        self.close()

    def cancelClicked(self, element):
        self.result = DialogResult.Cancel
        self.value = None
        self.close()

class EntryDialog(DialogBox):
    def __init__(self, app, title, label, validator=None):
        DialogBox.__init__(self, app, Canvas(400, 230), title)

        labelFont = app.screenManager.fonts.bigMenuFont
        labelColour = app.theme.colours.dialogBoxTextColour
        btnFont = app.screenManager.fonts.bigMenuFont
        btnColour = app.theme.colours.dialogButtonColour
        highlightColour = app.theme.colours.black
        inputFont = app.screenManager.fonts.defaultTextBoxFont
        inputColour = app.theme.colours.grey

        self.inputBox = InputBox(app,
            Region(topleft=self.Relative(0.1, 0.4),
                bottomright=self.Relative(0.9, 0.6)),
            font=inputFont, colour=inputColour, onClick=self.setFocus,
            onEnter=self.okClicked, onEsc=self.cancelClicked,
            validator=validator)

        self.elements = [
            TextElement(app, label,
                labelFont, Location(self.Relative(0.1, 0.2), 'midleft'),
                labelColour),
            self.inputBox,

            TextButton(app,
                Location(self.Relative(0.3, 0.9), 'center'),
                'Ok', btnFont, btnColour, highlightColour,
                onClick=self.okClicked),
            TextButton(app,
                Location(self.Relative(0.7, 0.9), 'center'),
                'Cancel', btnFont, btnColour, highlightColour,
                onClick=self.cancelClicked),
        ]

        self.setFocus(self.inputBox)

    def okClicked(self, element):
        self.result = DialogResult.OK
        self.value = self.inputBox.value
        self.close()

    def cancelClicked(self, element):
        self.result = DialogResult.Cancel
        self.value = self.inputBox.value
        self.close()

class TeamNameBox(EntryDialog):
    def __init__(self, app):
        EntryDialog.__init__(self, app, 'Custom Team', 'Team name:')

class GameDurationBox(EntryDialog):
    def __init__(self, app):
        EntryDialog.__init__(self, app, 'Custom Duration', 'Duration (mins):',
                intValidator)

class NicknameBox(EntryDialog):
    def __init__(self, app):
        EntryDialog.__init__(self, app, 'Nickname', 'New nickname:')




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging
import struct

from trosnoth.utils.message import MessageBase as Message

log = logging.getLogger(__name__)


##################################
# Message-related errors
##################################

class NetworkMessageError(Exception):
    '''
    An attempt was made to process a badly formatted network message.
    '''


class MessageTypeError(NetworkMessageError):
    '''
    An attempt was made to build a NetworkMessage object from a string which
    does not represent a valid network message type.
    '''


class MessageContentsError(NetworkMessageError):
    '''
    An attempt was made to build a NetworkMessage object from a badly formatted
    message string.
    '''


##################################
# Network message metaclass
##################################

class NetworkMessageClass(type):
    '''
    Metaclass which performs declaration-time checking of attributes of
    NetworkMessage subclasses.
    '''
    def __new__(cls, name, bases, dictn):
        # Test for a single string cls.fields and make tuple.
        if isinstance(dictn.get('fields'), str):
            dictn['fields'] = (dictn['fields'],)

        # Check for use of reserved names for fields.
        for field in dictn.get('fields', ()):
            if field in ('idString', 'fields', 'packspec'):
                raise NameError('Reserved field name: %r' % (field,))

        # Do the real building.
        result = type.__new__(cls, name, bases, dictn)

        if getattr(result, 'idString'):
            # Validate the class attributes.
            if not isinstance(result.idString, str):
                raise TypeError('Message type idString must be string')
            if len(result.idString) != 4:
                raise ValueError(
                    'id string %r does not have 4 characters' % (
                        result.idString,))

            # Check for the other 2 required attributes.
            if not hasattr(result, 'packspec'):
                raise AttributeError('packspec attribute required')
            if not hasattr(result, 'fields'):
                raise AttributeError('fields attribute required')

        return result


#############################
# Network message registries
#############################

class MessageCollection(object):
    def __init__(self, *messageTypes):
        self._msgTypes = {}
        for msgType in messageTypes:
            self._registerType(msgType)

    def __iter__(self):
        return self._msgTypes.itervalues()

    def _registerType(self, msgType):
        '''
        Registers the given network message type.
        '''
        if not isinstance(msgType.idString, str):
            raise TypeError('Message type idString must be string')
        if len(msgType.idString) != 4:
            raise ValueError('id string %r does not have 4 characters' % (
                msgType.idString,))
        if msgType.idString in self._msgTypes:
            raise KeyError('id string %r already used for %s' % (
                msgType.idString, self._msgTypes[msgType.idString]))

        self._msgTypes[msgType.idString] = msgType

    def buildMessage(self, input):
        '''
        Builds a NetworkMessage object from the packed string.
        '''
        if not isinstance(input, str):
            raise TypeError('%s is not a string' % (input,))

        # Look up the id string to find the message type.
        idString = input[:4]
        try:
            msgType = self._msgTypes[idString]
        except KeyError:
            raise MessageTypeError('bad id string: %r' % (idString))

        return msgType._build(input[4:])


############################
# Network message class
############################

class NetworkMessage(Message):
    '''
    Allows network messages to be defined by:
        class CustomNetMsg(NetworkMessage):
            fields = 'playerId', 'msgId', 'comment'
            packspec = 'cB*'
            idString = 'Cust'
    Parameters are as follows:
        fields - as with Message class
        packspec - a struct.pack string, except that network byte order will be
            automatically used (do not specify byte order), and the last
            character may optionally be * meaning that the rest of the message
            is considered to be a single string.
    '''
    __metaclass__ = NetworkMessageClass
    idString = None
    trace = False       # Set to True to see where a message goes

    def __init__(self, *args, **kwargs):
        # Check that all kwargs are valid keys.
        for k in kwargs.iterkeys():
            if k not in self.fields:
                raise TypeError('%s() got unexpected keyword argument %r' % (
                    self.__class__, k))

        # Initialise attributes based on arguments.
        Message.__init__(self, *args, **kwargs)

        # Check that packing can be performed.
        if self.idString:
            self.pack()

    def tracePoint(self, obj, method):
        if self.trace:
            log.warning('%s: %s (%s)', self, method, obj)

    def pack(self):
        '''
        Packs the field values of this message into a single string.
        '''
        # Collect the values to pack.
        values = []
        for k in self.fields:
            try:
                values.append(getattr(self, k))
            except AttributeError:
                raise ValueError('required field %s not provided' % (k,))

        # Special processing if there's a * in self.packspec.
        if self.packspec.endswith('*'):
            if not isinstance(values[-1], str):
                raise TypeError('%s field must be string' % (self.fields[-1]))
            suffix = values.pop()
            packspec = self.packspec[:-1]
        else:
            suffix = ''
            packspec = self.packspec

        # Perform the packing.
        try:
            result = struct.pack('!' + packspec, *values)
        except:
            raise Exception('%s message couldn\'t pack string %s' % (
                self.idString, str(values)))

        return self.idString + result + suffix

    @classmethod
    def _build(cls, source):
        '''
        Called by buildMessage() to build an instance of this class from a
        network message string.
        '''
        packspec = '!' + cls.packspec
        hasStar = packspec.endswith('*')
        if hasStar:
            packspec = packspec[:-1]
            size = struct.calcsize(packspec)
            suffix = source[size:]
            source = source[:size]
        else:
            # Verify size.
            size = struct.calcsize(packspec)
            if len(source) != size:
                raise MessageContentsError(
                    'bad message length for %s' % cls.__name__)

        # Do the actual unpacking.
        try:
            values = struct.unpack(packspec, source)
        except struct.error, E:
            raise MessageContentsError(
                'bad %s string: %s' % (cls.__name__, E.args[0]))

        # Build the message.
        if hasStar:
            values += (suffix,)
        return cls(*values)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging
from math import sin, cos, atan2, sqrt

from trosnoth.utils.collision import collideTrajectory
from trosnoth.utils.math import fadeValues

log = logging.getLogger(__name__)


class Unit(object):
    def __init__(self, world, *args, **kwargs):
        super(Unit, self).__init__(*args, **kwargs)
        self.world = world
        self.pos = (0, 0)
        self.oldPos = None

    def getMapBlock(self):
        return self.world.map.getMapBlockAtPoint(self.pos)

    def getZone(self):
        return self.world.map.getZoneAtPoint(self.pos)

    def isSolid(self):
        raise NotImplementedError

    def ignoreObstacle(self, obstacle):
        '''
        Returns True if this obstacle should be ignored in obstacle collision.
        Implemented by Player class for dropping through platforms.
        '''
        return False

    def continueOffMap(self):
        '''
        Called when this unit has fallen off the map. Returns True or False
        indicating whether this unit should keep falling or not.
        '''
        return True

    def canEnterZone(self, zone):
        '''
        Called when this unit attempts to enter the given zone. Should return
        True or False to indicate whether this zone entry is allowed.
        '''
        return True

    def reset(self):
        '''
        Called once every tick before any sprite has moved.
        '''
        self.oldPos = self.pos

    def tweenPos(self, fraction):
        if self.oldPos is None:
            return self.pos
        return (
            fadeValues(self.oldPos[0], self.pos[0], fraction),
            fadeValues(self.oldPos[1], self.pos[1], fraction),
        )

    @staticmethod
    def getObstacles(mapBlockDef):
        '''
        Return which obstacles in the given map block apply to this kind of
        unit.
        '''
        return mapBlockDef.obstacles


class CollectableUnit(Unit):
    '''
    Base class for units which have meaningful interactions with players.
    '''
    playerCollisionTolerance = 30

    def __init__(self, *args, **kwargs):
        super(CollectableUnit, self).__init__(*args, **kwargs)
        self.history = []
        self.hitLocalPlayer = False
        self.deathPeriod = 0

    def reset(self):
        if self.world.isServer:
            self.history.append(self.pos)
        super(CollectableUnit, self).reset()

    def beat(self):
        '''
        Called once per tick but only after this unit has been removed from the
        game for some reason.
        '''
        self.deathPeriod += 1

    def clearOldHistory(self, delay):
        delay -= self.deathPeriod
        self.history[:-delay - 1] = []

    def checkCollision(self, player, delay):
        '''
        Returns True if this unit has collided with the given player, based on
        where this unit was delay frames ago.
        '''
        delay -= self.deathPeriod
        if delay == 0:
            pos = self.pos
            if self.oldPos is None:
                oldPos = self.pos
            else:
                oldPos = self.oldPos
        else:
            if delay >= len(self.history) or delay < 0:
                return False
            pos = self.history[-delay]
            oldPos = self.history[-delay - 1]

        # Check both player colliding with us and us colliding with player
        if collideTrajectory(
                player.pos, oldPos, (pos[0] - oldPos[0], pos[1] - oldPos[1]),
                self.playerCollisionTolerance):
            return True

        deltaX = player.pos[0] - player.oldPos[0]
        deltaY = player.pos[1] - player.oldPos[1]
        if collideTrajectory(
                pos, player.oldPos, (deltaX, deltaY),
                self.playerCollisionTolerance):
            return True
        return False

    def collidedWithPlayer(self, player):
        '''
        Called when this unit has collided with the given player. Note that
        this is only calculated on the server, so anything done here will not
        automatically happen on clients too. Therefore, it's a good idea to
        just send a message, and let the message handler do the hard work here.
        '''
        raise NotImplementedError(
            '%s.collidedWithPlayer' % (self.__class__.__name__,))

    def collidedWithLocalPlayer(self, player):
        '''
        Called when this unit has collided with the given local player.
        '''
        self.hitLocalPlayer = True


class Bouncy(Unit):
    stopToleranceDistance = 1
    stopToleranceTicks = 5

    def __init__(self, world, *args, **kwargs):
        super(Bouncy, self).__init__(world, *args, **kwargs)
        self.stopped = False
        self.stationaryTicks = 0

    def isSolid(self):
        return True

    def advance(self):
        '''
        Called when this unit should update its position.
        deltaT is the time that's passed since its state was current, measured
        in seconds.
        '''
        if self.stopped:
            return

        deltaT = self.world.tickPeriod

        try:
            deltaX = self.xVel * deltaT
            deltaY = self.yVel * deltaT

            oldX, oldY = self.pos
            obstacle = self.world.physics.moveUnit(self, deltaX, deltaY)
            if ((self.pos[0] - oldX) ** 2 + (self.pos[1] - oldY) ** 2 <
                    self.stopToleranceDistance ** 2):
                self.stationaryTicks += 1
                if self.stationaryTicks > self.stopToleranceTicks:
                    self.stopped = True
                    self.xVel = self.yVel = 0
                    return
            else:
                self.stationaryTicks = 0

            if obstacle is not None:
                self.performRebound(obstacle)

            # v = u + at
            vFinal = self.yVel + self.getGravity() * deltaT
            if vFinal > self.getMaxFallVel():
                # Hit terminal velocity. Fall has two sections.
                deltaY = (
                    deltaY + (self.getMaxFallVel() ** 2 - self.yVel ** 2)
                    / (2 * self.getGravity()) + self.getMaxFallVel() *
                    (deltaT - (self.getMaxFallVel() - self.yVel) /
                        self.getGravity()))
                self.yVel = self.getMaxFallVel()
            else:
                # Simple case: s=ut+0.5at**2
                deltaY = (
                    deltaY + self.yVel * deltaT + 0.5 * self.getGravity()
                    * deltaT ** 2)
                self.yVel = vFinal

        except Exception, e:
            log.exception(str(e))

    def getGravity(self):
        raise NotImplementedError()

    def getMaxFallVel(self):
        raise NotImplementedError()

    def performRebound(self, obstacle):
        obsAngle = obstacle.getAngle()
        shotAngle = atan2(self.yVel, self.xVel)
        dif = shotAngle - obsAngle
        final = obsAngle - dif
        speed = sqrt(self.xVel ** 2 + self.yVel ** 2) * 0.9
        self.xVel = speed * cos(final)
        self.yVel = speed * sin(final)


class PredictedBouncyTrajectory(Bouncy):
    '''
    The trajectory of a grenade, without the explodey bits
    '''
    def __init__(
            self, world, player, duration, initialVelocity, gravity,
            maxFallVelocity, *args, **kwargs):
        Bouncy.__init__(self, world, *args, **kwargs)
        self.duration = duration
        self.player = player
        self.gravity = gravity
        self.maxFallVel = maxFallVelocity
        self.initialVelocity = initialVelocity

    def getGravity(self):
        return self.gravity

    def getMaxFallVel(self):
        return self.maxFallVel

    def predictedTrajectoryPoints(self):
        self.stopped = False
        self.timeLeft = self.duration
        self.pos = self.oldPos = self.player.pos
        angle = self.player.angleFacing
        self.xVel = self.initialVelocity * sin(angle)
        self.yVel = -self.initialVelocity * cos(angle)
        while self.timeLeft > 0 and not self.stopped:
            yield self.pos
            self.advance()
            self.timeLeft -= self.world.tickPeriod



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

MAX_SHOTS = 1 << 32 - 1
MAX_STARS = 255


class IdManager(object):
    '''
    Designed to be plugged on the server side. Takes all messages passed in
    (usually generated by the validator and universe) and passes them on after
    assigning shot ids, player ids, etc.
    '''

    def __init__(self, world):
        super(IdManager, self).__init__()
        self.world = world
        self._usedPlayers = set()
        self._nextPlayer = 1
        self._usedShots = set()
        self._nextShot = 1
        self._usedStars = set()
        self._nextStar = 1
        world.onShotRemoved.addListener(self.shotRemoved)
        world.onCollectableStarRemoved.addListener(self.collectableStarRemoved)
        world.onPlayerRemoved.addListener(self.playerRemoved)

    def stop(self):
        self.world.onShotRemoved.removeListener(self.shotRemoved)
        self.world.onCollectableStarRemoved.removeListener(
            self.collectableStarRemoved)
        self.world.onPlayerRemoved.removeListener(self.playerRemoved)

    def newShotId(self):
        shotId = self._nextShot
        if len(self._usedShots) >= MAX_SHOTS:
            return  None    # Cannot create any more shots.
        while shotId in self._usedShots:
            shotId = shotId % MAX_SHOTS + 1
        self._usedShots.add(shotId)
        self._nextShot = shotId % MAX_SHOTS + 1
        return shotId

    def shotRemoved(self, shotId):
        self._usedShots.discard(shotId)

    def newStarId(self):
        starId = self._nextStar
        if len(self._usedStars) >= MAX_STARS:
            return None     # Cannot create any more stars.
        while starId in self._usedStars:
            starId = starId % MAX_STARS + 1
        self._usedStars.add(starId)
        self._nextStar = starId % MAX_STARS + 1
        return chr(starId)

    def collectableStarRemoved(self, starId):
        self._usedStars.remove(ord(starId))

    def newPlayerId(self):
        playerId = self._nextPlayer
        if len(self._usedPlayers) >= 255:
            # Cannot join: hard max on player ids.
            return None

        while playerId in self._usedPlayers:
            playerId = playerId % 255 + 1
        self._usedPlayers.add(playerId)
        self._nextPlayer = playerId % 255 + 1

        return chr(playerId)

    def playerRemoved(self, player, playerId):
        self._usedPlayers.discard(ord(playerId))



import inspect
import logging
import types
import weakref

log = logging.getLogger(__name__)


class Event(object):
    def __init__(self, listener=None):
        self.listeners = {}
        if listener is not None:
            self.addListener(listener)

    def addListener(self, obj):
        assert callable(obj)
        if isinstance(obj, types.MethodType):
            if getattr(obj.im_self, obj.im_func.__name__, None) == obj:
                obj = WeakMethod(self, obj)
            else:
                log.warning(
                    'Cannot create weak reference for %s.%s()',
                    obj.im_self, obj.im_func.__name__)
        self.listeners[obj] = obj

    def removeListener(self, obj):
        try:
            value = self.listeners.pop(obj)
        except KeyError:
            return
        if WeakMethod is None:
            # Final garbage collection during process termination
            return
        if isinstance(value, WeakMethod):
            value.done()

    def clear(self):
        self.listeners.clear()

    def execute(self, *args, **kwargs):
        for call in list(self.listeners):
            try:
                call(*args, **kwargs)
            except Exception:
                caller = inspect.currentframe().f_back
                log.exception(
                    'Error in event callback (from %s:%s)',
                    caller.f_code.co_filename, caller.f_lineno)

    __call__ = execute


class WeakMethod(object):
    def __init__(self, event, method):
        self.event = event
        self.obj = weakref.ref(method.im_self, self.collect)
        self.attr = method.im_func.__name__
        self._hash = hash(method)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, (types.MethodType, WeakMethod)):
            return False
        return other == self.get_method()

    def get_method(self):
        if self.obj is None:
            return None
        obj = self.obj()
        if obj is None:
            return None
        return getattr(obj, self.attr)

    def __call__(self, *args, **kwargs):
        method = self.get_method()
        if method:
            method(*args, **kwargs)

    def collect(self, r):
        if self.obj is None:
            # This is possible if garbage collection has been delayed
            return
        self.event.removeListener(self)

    def done(self):
        # Lose the reference to the weak reference so that it can be garbage
        # collected and so that it doesn't keep a circular reference to
        # self.collect.
        self.obj = None



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

from trosnoth.model.mapLayout import MapLayout
from trosnoth.model.modes import PhysicsConstants

log = logging.getLogger(__name__)


class WorldPhysics(PhysicsConstants):
    def __init__(self, world):
        PhysicsConstants.__init__(self)
        self.world = world

    def moveUnit(self, unit, deltaX, deltaY, ignoreObstacles=()):
        '''
        Attempts to move the unit by the specified amount, taking into
        account the positions of walls. Also checks if the unit
        changes zones or changes map blocks.

        If the unit hit an obstacle, returns the obstacle.
        This routine only checks for obstacle collisions if unit.isSolid() is
        True.
        '''

        if unit.isSolid():
            lastObstacle, deltaX, deltaY = self.trimPathToObstacle(
                unit, deltaX, deltaY, ignoreObstacles)
        else:
            lastObstacle = None

        newX, newY = unit.pos[0] + deltaX, unit.pos[1] + deltaY
        i, j = MapLayout.getMapBlockIndices(newX, newY)
        zoneBlocks = self.world.zoneBlocks
        if i < 0 or j < 0 or i >= len(zoneBlocks) or j >= len(zoneBlocks[0]):
            if not unit.continueOffMap():
                return
        else:
            # Check for change of zones.
            newBlock = zoneBlocks[i][j]
            newZone = newBlock.getZoneAtPoint(newX, newY)
            if not unit.canEnterZone(newZone):
                return lastObstacle

        # Move the unit.
        unit.pos = (newX, newY)

        return lastObstacle

    def getNearbyObstacles(self, unit, deltaX, deltaY):
        '''
        Returns a superset of the obstacles which the given unit may collide
        with on the given path.
        '''
        def getObstacles(block):
            if block is None:
                return []
            return unit.getObstacles(block.defn)

        try:
            block = unit.getMapBlock()
        except IndexError:
            return []

        result = list(getObstacles(block))
        if deltaX < 0 and unit.pos[0] - block.defn.rect.left + deltaX < 0:
            result.extend(getObstacles(block.blockLeft))
        elif deltaX > 0 and unit.pos[0] - block.defn.rect.right + deltaX > 0:
            result.extend(getObstacles(block.blockRight))

        if deltaY < 0 and unit.pos[1] - block.defn.rect.top + deltaY < 0:
            result.extend(getObstacles(block.blockAbove))
        elif deltaY > 0 and unit.pos[1] - block.defn.rect.bottom + deltaY > 0:
            result.extend(getObstacles(block.blockBelow))

        return result

    def trimPathToObstacle(self, unit, deltaX, deltaY, ignoreObstacles):
        '''
        Takes the given trajectory and trims it down to the location where it
        collides with the nearest obstacle. Returns (obstacle, dx, dy) where
        obstacle is the obstacle hit or None, and (dx, dy) is the new
        trajectory.
        '''
        lastObstacle = None
        # Check for collisions with obstacles - find the closest obstacle.
        for obstacle in self.getNearbyObstacles(unit, deltaX, deltaY):
            if (unit.ignoreObstacle(obstacle) or obstacle in ignoreObstacles):
                continue
            dX, dY = obstacle.collide(unit, deltaX, deltaY)
            if (dX, dY) != (deltaX, deltaY):
                # Remember the last obstacle we hit.
                lastObstacle = obstacle
                deltaX = dX
                deltaY = dY
            if deltaX == 0 == deltaY:
                break

        return lastObstacle, deltaX, deltaY



import logging

from twisted.internet import defer

from trosnoth.messages import SetAgentPlayerMsg, ConnectionLostMsg
from trosnoth.model.agent import Agent
from trosnoth.model.hub import Hub, MAX_AGENT_ID, UnableToConnect

log = logging.getLogger(__name__)


class LocalHub(Hub):
    '''
    A Hub that connects to a local game by creating agents (rather than
    connecting to a remote game over the network).

    (In theory, this hub can even connect to a RemoteGame, but that is unlikely
    to be a useful state of affairs, since that ReboteGame will require another
    hub to connect to the network.)
    '''

    def __init__(self, game, *args, **kwargs):
        super(LocalHub, self).__init__(*args, **kwargs)
        self.game = game
        self.agents = {}
        self.nextId = 0

    def disconnectNode(self):
        super(LocalHub, self).disconnectNode()
        for agentId in self.agents.keys():
            self.disconnectAgent(agentId)

    @defer.inlineCallbacks
    def connectNewAgent(self):
        result = self.nextId
        if len(self.agents) >= MAX_AGENT_ID:
            raise UnableToConnect('No spare agent IDs')
        while result in self.agents:
            result = (result + 1) % (MAX_AGENT_ID + 1)

        agent = LocalHubAgent(self, result)
        yield self.game.addAgent(agent)

        self.agents[result] = agent
        self.nextId = (result + 1) % (MAX_AGENT_ID + 1)
        defer.returnValue(result)

    def disconnectAgent(self, agentId):
        agent = self.agents[agentId]
        self.game.detachAgent(agent)
        self.agentDisconnected(agent)

    def sendRequestToGame(self, agentId, msg):
        msg.tracePoint(self, 'sendRequestToGame')
        self.game.agentRequest(self.agents[agentId], msg)

    def agentDisconnected(self, agent):
        '''
        Called when the game indicates that the connection to this agent is
        lost.
        '''
        if agent.agentId not in self.agents:
            # It may be possible for this to be triggered twice for the same
            # agent, but we only want it to run once.
            return

        if self.node:
            self.node.agentDisconnected(agent.agentId)
        agent.stop()
        del self.agents[agent.agentId]


class LocalHubAgent(Agent):

    def __init__(self, hub, agentId, *args, **kwargs):
        super(LocalHubAgent, self).__init__(game=hub.game, *args, **kwargs)
        self.hub = hub
        self.agentId = agentId

    def messageToAgent(self, msg):
        msg.tracePoint(self, 'messageToAgent')
        if self.hub.node:
            self.hub.node.gotMessageToAgent(self.agentId, msg)

    def gotServerCommand(self, msg):
        msg.tracePoint(self, 'gotServerCommand')
        if self.hub.node:
            if isinstance(msg, ConnectionLostMsg):
                self.hub.agentDisconnected(self)
            else:
                self.hub.node.gotServerCommand(msg)

    def setPlayer(self, player):
        super(LocalHubAgent, self).setPlayer(player)

        # To get this across the hub/node interface it must be encoded as a
        # message.
        playerId = player.id if player else '\x00'
        self.messageToAgent(SetAgentPlayerMsg(playerId))



import logging

log = logging.getLogger(__name__)


enabled = False


# Treat every incoming instruction as if we've observed a delay of at least
# this number of ticks.
forceDelay = 0

# Slows down the universe game loop by this factor
slowMotionFactor = 1

# Maximum number of shots allowed on the map - for debugging shots
# (ignored if falsey)
shotLimit = None

# Shows AI path finding working
showPathFinding = True


# Displays where the local server thinks sprites are
showSpriteCircles = False


def getSpriteCircles():
    '''
    Returns a sequence of (point, radius) pairs for positions in the server
    universe where sprites are.
    '''
    if serverUniverse is None:
        return
    for unit in serverUniverse.getCollectableUnits():
        if localPlayerDelay == 0:
            pos = unit.pos
        else:
            if localPlayerDelay >= len(unit.history) or localPlayerDelay < 0:
                continue
            pos = unit.history[-localPlayerDelay]
        yield (pos, unit.playerCollisionTolerance)

    for player in serverUniverse.playerWithId.itervalues():
        yield (player.pos, 3)


# Updated by other parts of the code
localPlayerDelay = 0
localPlayerId = None
serverUniverse = None
